"""Tooling to attempt to flag components related to bandpasses. The idea
being that poorly calibration channels in the antenna-based gains should
be removed.

At this point there are no attempts to smooth or interpolate these flagged
components of the bandpass.
"""

from __future__ import annotations

from pathlib import Path
from typing import NamedTuple

import matplotlib.pyplot as plt
import numpy as np
from astropy.stats import sigma_clipped_stats
from scipy.optimize import curve_fit

from flint.exceptions import PhaseOutlierFitError
from flint.logging import logger


class PhaseOutlierResults(NamedTuple):
    """Results from the attempt to identify outlier complex gains in
    the bandpass solutions. This procedure is concerned with identifying
    channel-wise outliers by first unwrapping any uncorrected delay term
    in the complex_gains, fitting to the unwrapped phase components, and
    then fitting.
    """

    complex_gains: np.ndarray
    """The input gains to plot"""
    init_model_gains: np.ndarray
    """The initial model of the complex_gains"""
    fit_model_gains: np.ndarray
    """The complex gain model fit made against the unwrapped gains (i.e. complex_gains / init_model_gains)"""
    init_model_params: tuple[float, float]
    """The initial guess (gradient, offset) model parameters to represent the phase component of the complex_gains"""
    fit_model_params: tuple[float, float]
    """The fitted model parameters constrained against the unwrapped gains"""
    outlier_mask: np.ndarray
    """Boolean mask of equal length to complex_gain, where True represents outliers that should be flagged"""
    unwrapped_residual_mean: float
    """The mean of the residual unwrapped phases in radians"""
    unwrapped_residual_std: float
    """The std. of the residual unwrapped phases in radians"""
    flag_cut: float
    """The adopted significance level that a outlier should be before outlier_mask is set to True"""


# TODO: Pass in parameters directly so we don't have to have an instance of PhaseOutlierResults
def plot_phase_outlier(
    phase_outlier_results: PhaseOutlierResults,
    output_path: Path,
    title: str | None = None,
) -> Path:
    """Create a simple diagnostic plot highlighting how the outlier
    channels and their phases were selected.

    Args:
        phase_outlier_results (PhaseOutlierResults): Results from the outlier phase flagging method
        output_path (Path): Location to write the output plot to
        title (str, optional): Title to add to the figure. Defaults to None.

    Returns:
        Path: Path of the output image file
    """
    logger.debug(f"Creating phase outlier plot, writing {output_path!s}.")

    complex_gains = phase_outlier_results.complex_gains
    init_model_gains = phase_outlier_results.init_model_gains
    fit_model_gains = phase_outlier_results.fit_model_gains
    unwrapped_mean = phase_outlier_results.unwrapped_residual_mean
    unwrapped_std = phase_outlier_results.unwrapped_residual_std
    flag_cut = phase_outlier_results.flag_cut
    outlier_mask = phase_outlier_results.outlier_mask

    xs = np.arange(complex_gains.shape[0])

    residual_fit_gains = complex_gains / init_model_gains / fit_model_gains

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4))

    ax1.plot(xs, np.angle(complex_gains))
    ax1.plot(
        xs[outlier_mask], np.angle(complex_gains[outlier_mask]), "bo", label="Flagged"
    )
    ax1.plot(
        xs, np.angle(init_model_gains), color="red", label="Initial Model", alpha=0.4
    )
    ax1.plot(
        xs,
        np.angle(init_model_gains * fit_model_gains),
        color="black",
        label="Fitted Model",
        alpha=0.4,
    )
    ax1.legend()
    ax1.set(xlabel="Channels", ylabel="Phase (rad)", title="Raw Data")
    ax1.grid()

    ax2.plot(xs, np.angle(residual_fit_gains), label="Residual")
    ax2.plot(
        xs,
        np.angle(complex_gains / init_model_gains),
        label="Init. Model Unwrapped",
        alpha=0.4,
    )
    ax2.plot(
        xs[outlier_mask],
        np.angle(residual_fit_gains[outlier_mask]),
        "bo",
        label="Flagged",
    )
    ax2.axhline(unwrapped_mean, color="red", ls="-", label="Mean")
    ax2.axhline(
        unwrapped_mean - flag_cut * unwrapped_std,
        color="red",
        ls="--",
        label=f"+/- {flag_cut:.2f}sigma",
    )
    ax2.axhline(
        unwrapped_mean + flag_cut * unwrapped_std,
        color="red",
        ls="--",
    )

    ax2.grid()
    ax2.legend(ncol=2)
    ax2.set(xlabel="Channels", ylabel="Phase (rad)", title="Initial Unwrapped")

    if title:
        fig.suptitle(title)

    fig.tight_layout()

    fig.savefig(str(output_path))

    return output_path


def complex_gain_model(
    xs: np.ndarray, gradient: float, phase_offset: float
) -> np.ndarray:
    """Simulate a simple set of complex gains. No consideration made to amplitudes,
    only considering phase

    Args:
        xs (np.ndarray): Positions to evaluate model at.
        gradient (float): Gradient of the phase-slope, rate of wrapping
        phase_offset (float): Initial phasor offset of the model

    Returns:
        np.ndarray: Equal length to input xs of complex numbers representing phase-ramp
    """

    gains = np.exp(1j * 2.0 * np.pi * gradient * xs + phase_offset * 1j)
    return gains


def fit_complex_gain_model(*args):
    """A fudge to help the curve_fit along."""
    return np.angle(complex_gain_model(*args))


def flag_outlier_phase(
    complex_gains: np.ndarray,
    flag_cut: float,
    use_mad: bool = False,
    plot_title: str | None = None,
    plot_path: Path | None = None,
) -> PhaseOutlierResults:
    """This procedure attempts to identify channels in the bandpass solutions to
    flag but searching for gains with outlier phases. Typically, ASKAP solutions
    have a phase-slope across the band (i.e. a delay). This phase-slope first
    needs to be 'unwrapped' in order to correctly identify outliers reliably.

    Internally this function constructs an initial model of the unknown phase-slope.
    It estimates this by looking at the bulk set of gradients on a channel-to-channel
    basis, and looking at the builk distribution after removing large jumps (possibly
    RFI, possible a wrap). The initial phase offset is taken as the phase of the first
    valid gain.

    The initial model is used to unwrap the data, allowing a lest-squares fitter to
    more reliably fit. Once the fitter has been executed, the final cuts are applied
    against the unwrapped phase residuals.

    Experience shows that best results are achieved when the input complex-gains
    have been normalised against a reference antenna. There may be complex structures
    when the raw antenna phase vs frequency plots that can not be reliably fit for
    in this manner. These structures (I believe) arise from the beam-wise spectral
    sub-windows. See BPTools for a more thorough explanation.

    Args:
        complex_gains (np.ndarray): The comples-gains as a function of frequency.
        flag_cut (float): The significance a point should be before flagged as outlier
        use_mad (bool, optional): Use the median and MAD when selecting outlier. if False, use mean and std. Defaults to False.
        plot_title (str, optional): Title to add to the plot. Defaults to None.
        plot_path (Path, optional): If not None, a simple diagnostic plot will be created and written to this path. Defaults to None.

    Returns:
        PhaseOUtlierResults: Collection of results from this phase outlier flagging routine
    """

    idxs = np.arange(complex_gains.shape[0])

    # Step one: attempt to guess initial conditions of model to unwrap.
    # These calibration solutions typically have a unknown delay and phase
    # offset. For the least-squares fitter to fit the data robustly, decent
    # initial guesses are needed.
    complex_mask = np.isfinite(complex_gains)
    gain_angles = np.angle(complex_gains)
    init_phase_offset = gain_angles[complex_mask][0]

    # dividing by run is not strictly needed at the moment
    init_gradients = (gain_angles[1:] - gain_angles[:-1]) / (idxs[1:] - idxs[:-1])
    # The second half of this mask is to capture and exclude moments where the phase slop wraps.
    # The initial guess just needs to be in the ball park.
    init_gradients_mask = np.isfinite(init_gradients) & (
        np.abs(init_gradients) < np.pi / 2
    )
    init_gradient = np.median(init_gradients[init_gradients_mask])
    # TODO: Pretty sure this 2pi factor can be removed if the complex_gains_model
    # also has its 2pi removed. It feels like I am missing a something here.
    init_p0 = (init_gradient / (2.0 * np.pi), init_phase_offset)
    init_model_gains = complex_gain_model(idxs, *init_p0)

    # Now construct the initial guess model, used to unwrap the data
    unwrapped_complex_gains = complex_gains / init_model_gains
    unwrapped_complex_mask = np.isfinite(unwrapped_complex_gains)
    if np.sum(unwrapped_complex_mask) == 0:
        # No good measurements.
        raise PhaseOutlierFitError

    # Since there should be a fairly decent initial unwrapped with
    # an additional additive offset to set the bulk of the phases
    # to near zero, we can pass the fitter a fairly simple guesses
    p0 = [0, 0]
    try:
        results = curve_fit(
            fit_complex_gain_model,
            idxs[unwrapped_complex_mask],
            np.angle(unwrapped_complex_gains)[unwrapped_complex_mask],
            p0,
        )
    except ValueError:
        logger.critical("The phase outlier fit method failed. Flagging antenna. ")
        raise PhaseOutlierFitError

    fit_model_gains = complex_gain_model(idxs, *results[0])

    # Make the residuals
    unwrapped_residuals = np.angle(unwrapped_complex_gains / fit_model_gains)
    unwrapped_residuals_mask = np.isfinite(unwrapped_residuals)

    # Apply the final cuts to identify channels of excess phase offset, indicating
    # RFI.
    valid_residuals = unwrapped_residuals[unwrapped_residuals_mask]

    unwrapped_residual_median = np.median(valid_residuals)
    unwrapped_residual_mad = np.median(
        np.abs(valid_residuals - unwrapped_residual_median)
    )

    unwrapped_residual_mean = np.nanmean(unwrapped_residuals)
    unwrapped_residual_std = np.nanstd(unwrapped_residuals)

    if use_mad:
        m, s = unwrapped_residual_median, unwrapped_residual_mad
    else:
        m, s = unwrapped_residual_mean, unwrapped_residual_std

    final_mask = np.isfinite(unwrapped_residuals) & (
        np.abs(unwrapped_residuals) < (m + flag_cut * s)
    )

    phase_outlier_results = PhaseOutlierResults(
        complex_gains=complex_gains,
        init_model_gains=init_model_gains,
        fit_model_gains=fit_model_gains,
        init_model_params=init_p0,
        fit_model_params=results[0],
        outlier_mask=~final_mask,
        unwrapped_residual_mean=m,
        unwrapped_residual_std=s,
        flag_cut=flag_cut,
    )

    if plot_path:
        plot_phase_outlier(
            phase_outlier_results=phase_outlier_results,
            output_path=plot_path,
            title=plot_title,
        )

    return phase_outlier_results


def flags_over_threshold(
    flags: np.ndarray, thresh: float = 0.8, ant_idx: int | None = None
) -> bool:
    """Given a set of flags for an antenna across frequency, consider how much is flagged, indicated
    by a value of True, and return whether it was over a threshold. The intent is to return whether
    an entire antenna should be flagged.

    Args:
        flags (np.ndarray): Array of flags to consider
        thresh (float, optional): Threshold as a fraction that has to be meet before considered bad. Defaults to 0.8.
        ant_idx (Optional[int], optional): Number of the antenna being considered. Defaults to None.

    Returns:
        bool: Whether the number of flags has reached a threshold
    """

    assert 0.0 <= thresh <= 1.0, (
        f"The provided {thresh=} should be a fraction between 0 to 1. "
    )

    number_flagged = np.sum(flags)
    # Use the shape in case multi-dimensional array passed in
    total_flagged = np.prod(flags.shape)

    frac_flagged = number_flagged / total_flagged
    thresh_str = f"Total flagged: {frac_flagged:2.2f}"
    thresh_str = f"Antenna {ant_idx:02d} - {thresh_str}"

    logger.info(thresh_str)

    return frac_flagged > thresh


def plot_mean_amplitudes(
    amplitudes: np.ndarray,
    model: np.ndarray,
    mean: float,
    std: float,
    output_path: Path,
    plot_title: str | None = None,
) -> Path:
    """A simply plot to examine the polynomial fit to the residual amplitude data.

    Args:
        amplitudes (np.ndarray): The amplitudes being fit to
        model (np.ndarray): The evaluated polynomial that was fit to the data
        mean (float): The mean that was evaluated, which is used to plot onto the residual
        std (float): The deviation that was evaluated, which is used to highlight the dispersion on the residual
        output_path (Path): The location to save the output file to
        plot_title (Optional[str], optional): Title to add to the figure. Defaults to None.

    Returns:
        Path: Location the figure was saved to
    """

    fig, (ax1, ax2) = plt.subplots(1, 1)

    ax1.plot(amplitudes, color="blue", label="Data")
    ax1.plot(model, color="red", label="Model")
    ax1.legend()
    ax1.grid()
    ax1.set(xlabel="Channels", ylabel="Power")
    ax1.grid()

    ax2.plot(amplitudes - model, color="blue", label="Residual")
    ax2.axhline(mean, color="red", label=f"Mean - {mean:2.2f}")
    ax2.axhline(mean + std, color="red", ls="--", label=f"+/- 1sig: {std:2.2f}")
    ax2.grid()
    ax2.legend()
    ax2.set(xlabel="Channels", ylabel="Power")

    if plot_title:
        fig.suptitle(plot_title)

    fig.tight_layout()
    fig.savefig(str(output_path))

    return output_path


def flag_mean_residual_amplitude(
    complex_gains: np.ndarray,
    use_robust: bool = True,
    polynomial_order: int = 5,
    plot_path: Path | None = None,
    plot_title: str | None = None,
) -> bool:
    """Calculate the median or mean of the residual amplitudes of the complex gains
    after fitting a polynomial of order polynomial_order.

    If this median/mean is above 0.1 a value of True is returned, indicating
    that the antenna should be flagged.

    Internally this is fitting a simple polynomial to the amplitude spectrum. This method
    is not as robust as the moving polynomial harmonic fit as implemented in the bptools
    package. The method current;y implemented might not be robust to sudden changes in the
    solutions that might be present on the boundaries of beam forming intervals. In some
    instances this might be avoided or minimised if the input complex gains have been divided
    by a reference antenna.

    Args:
        complex_gains (np.ndarray): The set of complex gains to be considered
        use_robust (bool, optional): Whether to use robust statistics (median, MAD)  or mean/std to calculate the statistic against. Defaults to True.
        polynomial_order (int, optional): The order of the polynomial (numpy.polyfit) to use to compute the baseline. Defaults to 5.
        plot_path (Path): The location to save the output file to
        plot_title (Optional[str], optional): Title to add to the figure. Defaults to None.

    Returns:
        bool: Whether the data should be considered bad. True if it is bad, False if otherwise.
    """

    if not np.any(np.isfinite(complex_gains)):
        return True

    amplitudes = np.abs(complex_gains)
    idxs = np.arange(amplitudes.shape[0])
    mask = np.isfinite(amplitudes)

    poly_coeffs = np.polyfit(idxs[mask], amplitudes[mask], deg=polynomial_order)
    poly_vals = np.polyval(poly_coeffs, idxs)

    residual = amplitudes - poly_vals
    # Although the mask above should be sufficient, trust nothing
    mask = np.isfinite(residual)

    # TODO: Consider use of an iterative clipping method
    if use_robust:
        mean = np.median(residual[mask])
        deviation = np.abs(residual) - mean
        deviation = np.median(deviation[np.isfinite(deviation)])
    else:
        mean = np.mean(residual[mask])
        deviation = np.std(residual[mask])

    bad = np.abs(mean) > 0.1 or deviation > 0.5

    if plot_path:
        # TODO: Do this really belong here? It would perhaps be better to return a struct
        # so that all telescope polarisations could be plotted together.
        plot_mean_amplitudes(
            amplitudes=amplitudes,
            model=poly_vals,
            mean=mean,
            std=deviation,
            plot_title=plot_title,
            output_path=plot_path,
        )

    return bad


def flag_mean_xxyy_amplitude_ratio(
    xx_complex_gains: np.ndarray, yy_complex_gains, tolerance: float = 0.1
) -> bool:
    """Will robust compute through an iterative sigma-clipping procedure the
    mean XX and YY gain amplitudes. The ratio of these  means are computed,
    and if the are:

    >>> ratio < (1. / fraction) or ratio > fraction

    the data are considered bad and a `True` is returned (indicating that they
    should be flagged).

    Args:
        xx_complex_gains (np.ndarray): The XX complex gains to be considered
        yy_complex_gains (_type_): The YY complex gains to be considered
        tolerance (float, optional): The tolerance used used to distinguish a critical mean ratio threshold. Defaults to 0.10.

    Returns:
        bool: Whether data should be flagged (True) or not (False)
    """

    assert xx_complex_gains.shape == yy_complex_gains.shape, (
        f"Input xx and yy shapes do not match. {xx_complex_gains.shape=} {yy_complex_gains.shape=}"
    )
    logger.info("Calculating mean ratios. ")

    xx_amplitudes = np.abs(xx_complex_gains)
    yy_amplitudes = np.abs(yy_complex_gains)

    xx_mean, xx_median, xx_std = sigma_clipped_stats(
        data=xx_amplitudes, mask=~np.isfinite(xx_amplitudes)
    )
    yy_mean, yy_median, yy_std = sigma_clipped_stats(
        data=yy_amplitudes, mask=~np.isfinite(yy_amplitudes)
    )

    mean_gain_ratio = xx_mean / yy_mean

    result = (
        not np.isfinite(mean_gain_ratio)
        or mean_gain_ratio < (1.0 - tolerance)
        or mean_gain_ratio > (1.0 + tolerance)
    )

    if result:
        logger.warning(
            f"Failed the mean gain ratio test: {xx_mean=} {yy_mean=} {mean_gain_ratio=} {tolerance=}"
        )

    return result


def construct_mesh_ant_flags(mask: np.ndarray) -> np.ndarray:
    """Construct a mask that will accumulate the flags across
    all antennas. The input mask array should be boolean and
    of shape (ant, channels, pol), where `True` means flagged.

    If an antenna is completely flagged it is ignored as the
    statistics are collected

    Args:
        mask (np.ndarray): Input array denoting which items are flagged.

    Returns:
        np.ndarray: Output array where antennas have common sets of flags
    """

    assert len(mask.shape) == 3, (
        f"Expect array of shape (ant, channel, pol), received {mask.shape=}"
    )
    accumulate_mask = np.zeros_like(mask[0], dtype=bool)

    nant = mask.shape[0]
    logger.info(f"Accumulating flagged channels over {nant=} antenna")

    empty_ants: list[int] = []

    # TODO: This can be replaced with numpy broadcasting

    for ant in range(nant):
        ant_mask = mask[ant]
        if np.all(ant_mask):
            empty_ants.append(ant)
            continue

        accumulate_mask = accumulate_mask | ant_mask

    logger.info(f"Flags in accumulated mask: {np.sum(accumulate_mask)}")

    result_mask = np.zeros_like(mask, dtype=bool)

    for ant in range(nant):
        if ant in empty_ants:
            result_mask[ant, :, :] = True
        else:
            result_mask[ant] = accumulate_mask

    return result_mask


def construct_jones_over_max_amp_flags(
    complex_gains: np.ndarray, max_amplitude: float
) -> np.ndarray:
    """Construct and return a mask that would flag an entire Jones
    should there be an element whose amplitude is above a flagging
    threshold

    Args:
        complex_gains (np.ndarray): Complex gains that will have a mask constructed
        max_amplitude (float): The flagging threshold, any Jones with a member above this will be flagged

    Returns:
        np.ndarray: Boolean array of equal shape to `complex_gains`, with `True` indicating a flag
    """

    assert complex_gains.shape[-1] == 4, (
        f"Expected last dimension to be length 4, received {complex_gains.shape=}"
    )

    logger.info(f"Creating mask for Jones with amplitudes of {max_amplitude=}")
    complex_gains = complex_gains.copy()

    original_shape = complex_gains.shape

    # Calculate tehe amplitudes of each of the complex numbers
    # and construct the initial mask
    amplitudes = np.abs(complex_gains)
    mask = amplitudes > max_amplitude

    # Compress all but the last dimension into a single
    # rank so we can easily broadcast over
    mask = mask.reshape((-1, 4))

    # Now broadcast like a pirate
    flag_jones = np.any(mask, axis=1)
    mask[flag_jones, :] = True

    # Convert back to original shape
    mask = mask.reshape(original_shape)

    no_flagged = np.sum(mask)

    if no_flagged > 0:
        logger.warning(f"{no_flagged} items flagged with {max_amplitude=}")

    return mask



"""Basic tests for utility functions"""

from __future__ import annotations

import math
import os
import shutil
import time
from pathlib import Path
from typing import Any

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS

from flint.convol import BeamShape
from flint.exceptions import TimeLimitException
from flint.logging import logger
from flint.utils import (
    SlurmInfo,
    copy_directory,
    estimate_skycoord_centre,
    flatten_items,
    generate_strict_stub_wcs_header,
    generate_stub_wcs_header,
    get_beam_shape,
    get_packaged_resource_path,
    get_pixels_per_beam,
    get_slurm_info,
    hold_then_move_into,
    log_job_environment,
    parse_environment_variables,
    temporarily_move_into,
    timelimit_on_context,
)


def test_flatten_items():
    """Flatten a list of items recursively"""
    items = [[1], 2, [[3]], [[4, [5]]]]
    flat = flatten_items(items=items)
    expected = [1, 2, 3, 4, 5]
    assert flat == expected

    items = [1, 2, 3, 4, 5]
    assert items == flatten_items(items=items)


def some_long_function(minimum_time=5):
    t1 = time.time()
    while time.time() - t1 < minimum_time:
        sum = 0
        for i in range(1000):
            sum = sum + 1
    print(f"Time taken: {time.time() - t1} seconds")


def test_timelimit_on_context():
    """Raise an error should a function take longer than expected to run"""
    with pytest.raises(TimeLimitException):
        with timelimit_on_context(timelimit_seconds=1):
            some_long_function(minimum_time=20)

    with timelimit_on_context(timelimit_seconds=5):
        some_long_function(minimum_time=1)

    # This should make sure that the signal is not raised after the context left
    some_long_function(minimum_time=5)


@pytest.fixture(scope="session", autouse=True)
def set_slurm_env():
    """Set up variables for a specific test"""
    os.environ["SLURM_JOB_ID"] = "12345"
    os.environ["SLURM_ARRAY_TASK_ID"] = "54321"


def test_get_slurm_info_with_values(set_slurm_env):
    """See if the slurm environment information handles things properly. There should
    be no slurm environemtn variables present most of the time"""

    slurm_info = get_slurm_info()
    assert isinstance(slurm_info, SlurmInfo)
    assert slurm_info.job_id == "12345"
    assert slurm_info.task_id == "54321"


def test_hold_then_move_same_folder(tmpdir):
    a = Path(tmpdir) / "Captain"

    with hold_then_move_into(hold_directory=a, move_directory=a) as example:
        assert a == example


def test_log_environment(set_slurm_env):
    slurm_info = log_job_environment()

    assert isinstance(slurm_info, SlurmInfo)
    assert slurm_info.job_id == "12345"
    assert slurm_info.task_id == "54321"


def test_hold_then_test_errors(tmpdir):
    """Make sure some basic error handling"""

    a = Path(tmpdir) / "Jack.txt"
    b = Path(tmpdir) / "Sparrow.txt"

    a.touch()
    b.touch()

    with pytest.raises(AssertionError):
        with hold_then_move_into(hold_directory=a, move_directory=b) as _:
            logger.info("This will not be here")


def test_hold_then_move_into_none(tmpdir: Any):
    """See whether the context manager behaves as expected when the temporary hold
    directory is None. This should just do thing in the move_directory."""

    tmpdir = Path(tmpdir)

    hold_directory = None
    move_directory = Path(tmpdir / "new/the/final/location")

    no_files = 45
    with hold_then_move_into(
        hold_directory=hold_directory, move_directory=move_directory
    ) as put_dir:
        assert put_dir.exists()
        assert put_dir == move_directory
        for i in range(no_files):
            file: Path = put_dir / f"some_file_{i}.txt"
            file.write_text(f"This is a file {i}")

        assert len(list(put_dir.glob("*"))) == no_files
        assert move_directory.exists()

    assert len(list(move_directory.glob("*"))) == no_files
    assert put_dir.exists()


def test_hold_then_move_into(tmpdir: Any):
    """See whether the hold directory can have things dumped into it, then
    moved into place on exit of the context manager"""

    tmpdir = Path(tmpdir)

    hold_directory = Path(tmpdir / "putthingshere")
    move_directory = Path(tmpdir / "the/final/location")

    assert all([not d.exists() for d in (hold_directory, move_directory)])
    no_files = 45
    with hold_then_move_into(
        hold_directory=hold_directory, move_directory=move_directory
    ) as put_dir:
        assert put_dir.exists()
        for i in range(no_files):
            file: Path = put_dir / f"some_file_{i}.txt"
            file.write_text(f"This is a file {i}")

        assert len(list(put_dir.glob("*"))) == no_files
        assert move_directory.exists()
        assert len(list(move_directory.glob("*"))) == 0

    assert len(list(move_directory.glob("*"))) == no_files
    assert not put_dir.exists()
    assert not hold_directory.exists()


def test_temporarily_move_into_none(tmpdir):
    """Make sure that the temporary context manager returns the same path without
    any deleting should the temporary directory be set to None"""

    TEXT = "this is a test message"
    source_test = Path(tmpdir) / "source_dir2/test.txt"
    source_test.parent.mkdir()
    source_test.touch()
    assert source_test.read_text() == ""

    with temporarily_move_into(
        subject=source_test, temporary_directory=None
    ) as temp_file:
        assert isinstance(temp_file, Path)
        assert source_test.samefile(temp_file)
        temp_file.write_text(TEXT)

    assert source_test.read_text() == TEXT


def test_temporarily_move_into_with_directory(tmpdir):
    """See whether the temp move context manager behaves in a sane way using the
    case where the subject is a directory"""
    TEXT = "this is a test message"
    source_test = Path(tmpdir) / "source_dir2/test.txt"
    source_test.parent.mkdir()
    source_test.touch()
    assert source_test.read_text() == ""
    source_parent = source_test.parent

    temp_dir = Path(tmpdir) / "someotherdirforwdir"
    assert not temp_dir.exists()

    with temporarily_move_into(
        subject=source_parent, temporary_directory=temp_dir
    ) as temp_parent:
        assert isinstance(temp_parent, Path)
        assert Path(temp_parent).exists()
        assert Path(temp_parent).is_dir()
        temp_file = Path(temp_parent) / "test.txt"
        assert temp_file.read_text() == ""

        temp_file.write_text(TEXT)
        assert temp_file.read_text() == TEXT

    assert source_test.read_text() == TEXT
    assert not temp_file.exists()
    assert not temp_dir.exists()


def test_temporarily_move_into(tmpdir):
    """See whether the temp move context manager behaves in a sane way"""
    TEXT = "this is a test message"
    source_test = Path(tmpdir) / "source_dir/test.txt"
    source_test.parent.mkdir()
    source_test.touch()
    assert source_test.read_text() == ""

    temp_dir = Path(tmpdir) / "someotherdir"
    assert not temp_dir.exists()

    with temporarily_move_into(
        subject=source_test, temporary_directory=temp_dir
    ) as temp_file:
        assert isinstance(temp_file, Path)
        assert Path(temp_file).exists()
        assert Path(temp_file).is_file()

        assert temp_file.read_text() == ""

        temp_file.write_text(TEXT)
        assert temp_file.read_text() == TEXT

    assert source_test.read_text() == TEXT
    assert not temp_file.exists()
    assert not temp_dir.exists()


@pytest.fixture(scope="session", autouse=True)
def set_env():
    """Set up variables for a specific test"""
    os.environ["TEST1"] = "Pirates"
    os.environ["TEST2"] = "Treasure"


def test_parse_environment_variables(set_env):
    """Make sure that the variable is processed nicely when getting environment variable"""
    val = parse_environment_variables("$TEST1")
    assert val == "Pirates"
    val2 = parse_environment_variables("$TEST2")
    assert val2 == "Treasure"
    val3 = parse_environment_variables("THISNOEXISTS")
    assert (
        val3 == "THISNOEXISTS"
    )  # in case the user gives a path that is not a variable return the same thing
    val4 = parse_environment_variables("$TEST1/$TEST2")
    assert val4 == "Pirates/Treasure"
    val5 = parse_environment_variables("$TEST1/$TEST2/TEST3")
    assert (
        val5 == "Pirates/Treasure/TEST3"
    )  # in case the user gives a path that includes variables and directories

    val6 = parse_environment_variables("$TEST1/$TEST2/$TEST3")
    assert val6 is None  # If one goes down all do
    val7 = parse_environment_variables("$TEST1/$TEST3/$TEST2")
    assert val7 is None  # Same as above but reordered

    # Same as above but with a trailing slash
    val8 = parse_environment_variables("$TEST1/$TEST2/")
    assert val8 == "Pirates/Treasure/"

    assert parse_environment_variables(variable=None) is None


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def test_copy_directory(ms_example, tmpdir):
    """See if we can copy folders"""
    out = Path(tmpdir) / "2"
    out.mkdir(exist_ok=True)
    out = out / ms_example.name

    copy_directory(input_directory=ms_example, output_directory=out)
    # Ensure overwrite works
    copy_directory(input_directory=ms_example, output_directory=out, overwrite=True)
    with pytest.raises(FileExistsError):
        copy_directory(input_directory=ms_example, output_directory=out)

    copy_directory(
        input_directory=ms_example, output_directory=out, overwrite=True, verify=True
    )


@pytest.fixture
def rms_path(tmpdir):
    rms_path = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0-MFS-subimage_rms.fits",
        )
    )

    return rms_path


def test_pixels_per_beam(rms_path):
    """Confirm pixels per beam is working"""
    no_pixels = get_pixels_per_beam(fits_path=rms_path)

    assert np.isclose(math.floor(no_pixels), 51.0)  # type: ignore
    assert no_pixels is not None
    assert no_pixels > 0.0


def test_get_beam_shape(rms_path):
    """Test to ensure the beam can be extracted from input image"""
    beam = get_beam_shape(fits_path=rms_path)

    assert isinstance(beam, BeamShape)
    assert np.isclose(10.90918, beam.bmaj_arcsec)
    assert np.isclose(9.346510, beam.bmin_arcsec)
    assert np.isclose(56.2253417, beam.bpa_deg)


def test_generate_strict_header_position():
    """Use a reference WCS to calculate the pixel position expected of a source"""

    # The following WCS are taken from SB40470, beam 17, and the header produced from wsclean
    # This beam has 3C 298 in it
    wcs_dict = dict(
        NAXIS1=8128,
        NAXIS2=8128,
        ORIGIN="WSClean",
        CTYPE1="RA---SIN",
        CRPIX1=4065,
        CRVAL1=-1.722664244157e02,
        CDELT1=-6.944444444444e-04,
        CUNIT1="deg",
        CTYPE2="DEC--SIN",
        CRPIX2=4065,
        CRVAL2=2.625771981318e00,
        CDELT2=6.944444444444e-04,
        CUNIT2="deg",
        CTYPE3="FREQ",
        CRPIX3=1,
        CRVAL3=8.874907407407e08,
        CDELT3=2.880000000000e08,
        CUNIT3="Hz",
        CTYPE4="STOKES",
        CRPIX4=1,
        CRVAL4=1.000000000000e00,
        CDELT4=1.000000000000e00,
        CUNIT4="",
    )
    center_position = SkyCoord(wcs_dict["CRVAL1"] * u.deg, wcs_dict["CRVAL2"] * u.deg)

    wcs = generate_strict_stub_wcs_header(
        position_at_image_center=center_position,
        image_shape=(
            int(wcs_dict["CRPIX1"]),
            int(wcs_dict["CRPIX2"]),
        ),
        pixel_scale=wcs_dict["CDELT2"] * u.deg,
        image_shape_is_center=True,
    )

    logger.info(f"{wcs=}")

    known_tato = SkyCoord("12:29:06 02:03:08", unit=(u.hourangle, u.deg))

    pixels = wcs.world_to_pixel(known_tato)
    logger.info(pixels)
    pixels = tuple([int(np.round(pixels[0])), int(np.round(pixels[1]))])

    assert pixels == (4724, 3238)


def test_generate_strict_wcs_header():
    """Generate an expects WCS header from known inputs"""
    image_shape = (2000, 2000)
    w = generate_strict_stub_wcs_header(
        position_at_image_center=SkyCoord(180, -30, unit=(u.deg, u.deg)),
        image_shape=image_shape,
        pixel_scale=-2.5 * u.arcsec,
    )
    assert isinstance(w, WCS)
    assert w.wcs.ctype[0] == "RA---SIN"
    assert w.wcs.ctype[1] == "DEC--SIN"

    w = generate_strict_stub_wcs_header(
        position_at_image_center=SkyCoord(180, -30, unit=(u.deg, u.deg)),
        image_shape=image_shape,
        pixel_scale="2.5arcsec",
    )
    assert isinstance(w, WCS)
    assert w.wcs.ctype[0] == "RA---SIN"
    assert w.wcs.ctype[1] == "DEC--SIN"


def test_wcs_getter():
    """Make a basic wcs object"""
    w = generate_stub_wcs_header(
        ra=180, dec=-45, image_shape=(8000, 8000), pixel_scale=0.01
    )

    assert isinstance(w, WCS)
    assert w._naxis == (8000, 8000)
    assert w.wcs.ctype[0] == "RA---SIN"
    assert w.wcs.ctype[1] == "DEC--SIN"


def test_wcs_getter_with_valueerrors(rms_path):
    """Make a basic wcs object"""
    # This one tests the pixel scale not being a quantity
    with pytest.raises(AssertionError):
        _ = generate_stub_wcs_header(ra=180, dec=-45, pixel_scale=2)

    # This one tests something being a None
    with pytest.raises(ValueError):
        _ = generate_stub_wcs_header(ra=180, dec=-45, pixel_scale="2.5arcsec")

    # This one tests a bad projection
    with pytest.raises(AssertionError):
        _ = generate_stub_wcs_header(
            ra=180, dec=-45, projection="ThisIsBad", pixel_scale="2.5arcsec"
        )

    # This one tests missing ra and dec but draws from the base
    _ = generate_stub_wcs_header(
        projection="SIN", pixel_scale="2.5arcsec", base_wcs=rms_path
    )

    # This one tests drawing everything from the base wcs
    w = generate_stub_wcs_header(base_wcs=rms_path)
    assert w._naxis == (15, 10)
    assert w.wcs.ctype[0] == "RA---SIN"
    assert w.wcs.ctype[1] == "DEC--SIN"

    w = generate_stub_wcs_header(
        ra=180, dec=-45, pixel_scale="2.5arcsec", base_wcs=rms_path
    )
    assert w._naxis == (15, 10)
    assert w.wcs.ctype[0] == "RA---SIN"
    assert w.wcs.ctype[1] == "DEC--SIN"


def test_wcs_getter_positions():
    """Make a basic wcs object"""
    # TODO: Need some proper tests here. Translate to sky positions etc
    w = generate_stub_wcs_header(
        ra=180, dec=-45, image_shape=(8000, 8000), pixel_scale=0.01
    )

    assert isinstance(w, WCS)

    w2 = generate_stub_wcs_header(
        ra=(180 * u.deg).to(u.rad),
        dec=-(45 * u.deg).to(u.rad),
        image_shape=(8000, 8000),
        pixel_scale=0.01,
    )

    assert isinstance(w, WCS)
    assert np.allclose(w.wcs.crval, w2.wcs.crval)


def test_wcs_getter_quantity():
    """Make a basic wcs object that includes different pixel types"""
    w = generate_stub_wcs_header(
        ra=180, dec=-45, image_shape=(8000, 8000), pixel_scale=0.01 * u.deg
    )
    assert isinstance(w, WCS)

    w = generate_stub_wcs_header(
        ra=180, dec=-45, image_shape=(8000, 8000), pixel_scale="2.5arcsec"
    )
    assert isinstance(w, WCS)

    w = generate_stub_wcs_header(
        ra=180 * u.deg,
        dec=-45 * u.rad,
        image_shape=(8000, 8000),
        pixel_scale="2.5arcsec",
    )
    assert isinstance(w, WCS)


def test_wcs_getter_withbase(rms_path):
    """Make a wcs object overriding the wcs from an existing fits file"""
    hdr = fits.getheader(rms_path)
    w = generate_stub_wcs_header(
        ra=180,
        dec=-45,
        image_shape=(8000, 8000),
        pixel_scale=0.01,
        base_wcs=WCS(hdr),
    )

    assert isinstance(w, WCS)

    w2 = generate_stub_wcs_header(
        ra=180,
        dec=-45,
        image_shape=(8000, 8000),
        pixel_scale=0.01,
        base_wcs=rms_path,
    )

    assert isinstance(w2, WCS)


def test_package_resource_path_folder():
    """Ensure the utils package path resource getter works"""
    dir_path = get_packaged_resource_path(package="flint.data", filename="")

    assert isinstance(dir_path, Path)
    assert dir_path.exists()


def test_package_resource_path_askap_lua():
    """Ensure the utils package path resource getter work, and check the contents of a file"""
    askap_lua = get_packaged_resource_path(
        package="flint.data.aoflagger", filename="ASKAP.lua"
    )

    assert isinstance(askap_lua, Path)
    assert askap_lua.exists()

    with open(askap_lua) as open_lua:
        line = open_lua.readline()
        assert line == "--[[\n"


def test_package_resource_path_skymodel():
    """Ensure the utils package path resource getter work, and check the contents of a file"""
    askap_model = get_packaged_resource_path(
        package="flint.data.models", filename="1934-638.calibrate.txt"
    )

    assert isinstance(askap_model, Path)
    assert askap_model.exists()

    with open(askap_model) as open_model:
        line = open_model.readline()
        assert (
            line
            == "Format = Name, Type, Ra, Dec, I, SpectralIndex, LogarithmicSI, ReferenceFrequency='888500000.0', MajorAxis, MinorAxis, Orientation\n"
        )


def test_estimate_skycoord_centre():
    """Estimate the centre position from a collection of sky positions"""
    ras = np.arange(-3, 3, 1) + 180.0
    decs = np.arange(-3, 3, 1) - 40.0

    sky_pos = SkyCoord(ras, decs, unit=(u.deg, u.deg))

    mean_pos = estimate_skycoord_centre(sky_positions=sky_pos)

    print(mean_pos)

    assert np.isclose(mean_pos.ra.deg, 179.54350474)
    assert np.isclose(mean_pos.dec.deg, -40.51256163)


def test_estimate_skycoord_centre_wrap():
    """Estimate the mean center sky position that wraps around 360 -> 0 degrees in ra"""
    ras = np.arange(-3, 3, 1) + 360.0 % 360
    decs = np.arange(-3, 3, 1) - 40.0

    sky_pos = SkyCoord(ras, decs, unit=(u.deg, u.deg))

    mean_pos = estimate_skycoord_centre(sky_positions=sky_pos)

    print(mean_pos)

    assert np.isclose(mean_pos.ra.deg, 359.54349533)
    assert np.isclose(mean_pos.dec.deg, -40.51255648)



from __future__ import annotations

import pytest

from flint.logging import logger

from .test_helpers import which

if which("singularity") is None:
    pytest.skip("Singularity is not installed", allow_module_level=True)


def test_singularity():
    which_singularity = which("singularity")
    logger.info(f"Singularity is installed at: {which_singularity}")
    assert which_singularity is not None



"""Operations related to measurement sets

Measurement sets are represented by the `MS` class. It describes
the path to a measurement set and any activate columns that should
be used for certain operations. Throughout `flint` steps are
generally carried out against the named column described by the
`column` attribute.
"""

from __future__ import annotations

import shutil
from argparse import ArgumentParser
from contextlib import contextmanager
from curses.ascii import controlnames
from os import PathLike
from pathlib import Path
from typing import NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from casacore.tables import table, taql
from fixms.fix_ms_corrs import fix_ms_corrs
from fixms.fix_ms_dir import fix_ms_dir

from flint.casa import copy_with_mstranform
from flint.logging import logger
from flint.naming import create_ms_name
from flint.options import MS
from flint.utils import copy_directory, rsync_copy_directory


class MSSummary(NamedTuple):
    """Small structure to contain overview of a MS"""

    unflagged: int
    """Number of unflagged records"""
    flagged: int
    """Number of flagged records"""
    flag_spectrum: np.ndarray
    """Flagged spectral channels"""
    fields: list[str]
    """Collection of unique field names from the FIELDS table"""
    ants: list[int]
    """Collection of unique antennas"""
    beam: int
    """The ASKAP beam number of the measurement set"""
    path: Path
    """Path to the measurement set that is being represented"""
    phase_dir: SkyCoord
    """The phase direction of the measurement set, which will be where the image will be centred"""
    spw: int | None = None
    """Intended to be used with ASKAP high-frequency resolution modes, where the MS is divided into SPWs"""


# TODO: Some common MS validation functions?
# - list / number of fields
# - new name function (using names / beams)
# - check to see if fix_ms_dir / fix_ms_corrs
# - delete column/rename column


@contextmanager
def critical_ms_interaction(
    input_ms: Path, copy: bool = False, suffix: str = ".critical"
):
    """A context manager that can be used to register a measurement set as
    entering a critical segment, i.e. phase rotation. If this stage were to
    fail, then future operations against said measurement set might be
    nonsense. This mechanism is intended to make it clear that the measurement
    set is in a dangerous part of code.

    Failure to return the MS to its original name (or rename the copy) highlights
    this failed stage.

    Args:
        input_ms (Path): The measurement set to monitor.
        copy (bool, optional): If True, a copy of the MS is made with the suffix supplied. Otherwise, the MS is siomply renamed. Defaults to False.
        suffix (str, optional): Suffix indicating the MS is in the dangerous stage. Defaults to '.critical'.

    Yields:
        Path: Resource location of the measurement set being processed
    """
    input_ms = Path(input_ms)
    output_ms: Path = Path(input_ms.with_suffix(suffix))

    # Make sure that the measurement set to preserve exists, and there is not already
    # a target output measurement set already on disk. This second check would be
    # useful in copy==True mode, ya seadog
    assert input_ms.exists(), f"The input measurement set {input_ms} does not exist. "
    assert not output_ms.exists(), (
        f"The output measurement set {output_ms} already exists. "
    )
    logger.info(f"Critical section for {input_ms=}")
    if copy:
        rsync_copy_directory(target_path=input_ms, out_path=output_ms)
    else:
        input_ms.rename(target=output_ms)

    try:
        yield output_ms
    except Exception as e:
        logger.error(
            f"An error occurred when interacting with {input_ms} during a critical stage. "
        )
        raise e

    # If we get to here, things worked successfully, and we
    # should return things back to normal.
    if copy:
        shutil.rmtree(input_ms)
        output_ms.rename(input_ms)
    else:
        output_ms.rename(target=input_ms)


def get_field_id_for_field(ms: MS | Path, field_name: str) -> int | None:
    """Return the FIELD_ID for an elected field in a measurement set

    Args:
        ms (Union[MS, Path]): The measurement set to inspect
        field_name (str): The desired name of the field to find the FIELD_ID for

    Raises:
        ValueError: Raised when more than one unique FIELD_ID is found for a single field

    Returns:
        Union[int, None]: The FIELD_ID as an `int` if the field is found. `None` if it is not found.
    """
    ms_path = ms if isinstance(ms, Path) else ms.path

    with table(f"{ms_path!s}/FIELD", readonly=True, ack=False) as tab:
        # The ID is _position_ of the matching row in the table.
        field_names = tab.getcol("NAME")
        field_idx = np.argwhere([fn == field_name for fn in field_names])[0]

        if len(field_idx) == 0:
            return None
        elif len(field_idx) > 1:
            raise ValueError(
                f"More than one matching field name found. This should not happen. {field_name=} {field_names=}"
            )

        field_idx = field_idx[0]
        logger.info(f"{field_name} FIELD_ID is {field_idx}")

    return field_idx


def get_beam_from_ms(ms: MS | Path) -> int:
    """Lookup the ASKAP beam number from a measurement set.

    Args:
        ms (Union[MS, Path]): The measurement set to inspect. If `MS`, the attached path is used.

    Returns:
        int: The beam ID number
    """
    ms_path = ms if isinstance(ms, Path) else ms.path

    with table(str(ms_path), readonly=True, ack=False) as tab:
        uniq_beams = sorted(list(set(tab.getcol("FEED1"))))

    assert len(uniq_beams) == 1, (
        f"Expected {ms_path!s} to contain a single beam, found {len(uniq_beams)}: {uniq_beams=}"
    )

    return uniq_beams[0]


def get_freqs_from_ms(ms: MS | Path) -> np.ndarray:
    """Return the frequencies observed from an ASKAP Measurement set.
    Some basic checks are performed to ensure they conform to some
    expectations.

    Args:
        ms (Union[MS, Path]): Measurement set to inspect

    Returns:
        np.ndarray: A squeeze array of frequencies, in Hertz.
    """
    ms = MS.cast(ms)

    with table(f"{ms.path!s}/SPECTRAL_WINDOW", readonly=True, ack=False) as tab:
        freqs = tab.getcol("CHAN_FREQ")

    freqs = np.squeeze(freqs)
    assert len(freqs.shape) == 1, (
        f"Frequency axis has dimensionality greater than one. Not expecting that. {len(freqs.shape)}"
    )

    return freqs


def get_phase_dir_from_ms(ms: MS | Path) -> SkyCoord:
    """Extract the phase direction from a measurement set.

    If more than one phase direction is found an AssertError will
    be raised.

    Args:
        ms (Union[MS, Path]): The measurement set to get the phase direction from

    Returns:
        SkyCoord: The phase direction the measurement set is directed towards
    """
    ms = MS.cast(ms)

    with table(f"{ms.path!s}/FIELD", readonly=True, ack=False) as tab:
        phase_dir = tab.getcol("PHASE_DIR")[0]

    assert phase_dir.shape[0] == 1, "More than one phase direction found. "

    phase_sky = SkyCoord(*phase_dir[0], unit=(u.rad, u.rad))

    return phase_sky


def get_times_from_ms(ms: MS | Path) -> Time:
    """Return the observation times from an ASKAP Measurement set.

    Args:
        ms (Union[MS, Path]): Measurement set to inspect

    Returns:
        Time: The observation times
    """
    # Get the time of the observation
    ms = MS.cast(ms)
    with table(str(ms.path), ack=False) as tab:
        times = Time(tab.getcol("TIME") * u.s, format="mjd")

    return times


def get_telescope_location_from_ms(ms: MS | Path) -> EarthLocation:
    """Return the telescope location from an ASKAP Measurement set.

    Args:
        ms (Union[MS, Path]): Measurement set to inspect

    Returns:
        EarthLocation: The telescope location
    """
    ms = MS.cast(ms)
    # Get the position of the observatory
    with table(str(ms.path / "ANTENNA"), ack=False) as tab:
        pos = EarthLocation.from_geocentric(
            *tab.getcol("POSITION")[0] * u.m  # First antenna is fine
        )
    return pos


def get_pol_axis_from_ms(
    ms: MS | Path, feed_idx: int | None = None, col: str = "RECEPTOR_ANGLE"
) -> u.Quantity:
    """Get the polarization axis from the ASKAP MS. Checks are performed
    to ensure this polarisation axis angle is constant throughout the observation.


    Args:
        ms (Union[MS, Path]): The path to the measurement set that will be inspected
        feed_idx (Optional[int], optional): Specify the entry in the FEED
        table of `ms` to return. This might be required when a subset of a
        measurement set has been extracted from an observation with a varying
        orientation.
        col (str, optional): The column to extract the polarization angle from.

    Returns:
        astropy.units.Quantity: The rotation of the PAF throughout the observing.
    """
    ms = MS.cast(ms=ms)

    # The INSTRUMENT_RECEPTOR_ANGLE is inserted from fixms.
    _known_cols = ("RECEPTOR_ANGLE", "INSTRUMENT_RECEPTOR_ANGLE")
    if col not in _known_cols:
        raise ValueError(f"Unknown column {col=}, please use one of {_known_cols}")

    with table((ms.path / "FEED").as_posix(), readonly=True, ack=False) as tf:
        if col not in tf.colnames():
            raise ValueError(f"{col=} not in the column names available. ")

        ms_feed = tf.getcol(col) * u.rad
        # PAF is at 45deg to feeds
        # 45 - feed_angle = pol_angle
        pol_axes = -(ms_feed - (45.0 * u.deg))  # type: ignore

    if feed_idx is None:
        assert (ms_feed[:, 0] == ms_feed[0, 0]).all() & (
            ms_feed[:, 1] == ms_feed[0, 1]
        ).all(), f"The {col} changes with time, please check the MS"

        feed_idx = 0

    logger.debug(f"Extracting the third-axis orientation for {feed_idx=}")
    pol_ang = pol_axes[feed_idx, 0].to(u.deg)

    assert pol_ang is not None, f"{pol_ang=}, which should not happen"
    return pol_ang


# TODO: Inline with other changing conventions this should be
# changed to `create_ms_summary`
def describe_ms(ms: MS | Path, verbose: bool = False) -> MSSummary:
    """Print some basic information from the inpute measurement set.

    Args:
        ms (Union[MS,Path]): Measurement set to inspect
        verbose (bool, optional): Log MS options to the flint logger. Defaults to False.

    Returns:
        MSSummary: Brief overview of the MS.

    """
    ms = MS(path=ms) if isinstance(ms, Path) else ms
    logger.info(f"Obtaining MSSummary for {ms.path}")

    with table(str(ms.path), readonly=True, ack=False) as tab:
        colnames = tab.colnames()

        flags: np.ndarray = tab.getcol("FLAG")
        flagged = np.sum(flags == True)  # Noqa: E712
        unflagged = np.sum(flags == False)  # Noqa: E712
        total = np.prod(flags.shape)
        flag_spectrum = flags.sum(axis=(0, -1)) / (flags.shape[0] * flags.shape[-1])

        uniq_ants = sorted(list(set(tab.getcol("ANTENNA1"))))

    with table(f"{ms.path}/FIELD", readonly=True, ack=False) as tab:
        uniq_fields = list(set(tab.getcol("NAME")))

    beam_no = get_beam_from_ms(ms=ms)
    phase_dir = get_phase_dir_from_ms(ms=ms)

    if verbose:
        logger.info(f"Inspecting {ms.path}.")
        logger.info(f"Contains: {colnames}")

        logger.info(f"{flagged} of {total} flagged ({flagged / total * 100.0:.4f}%). ")
        logger.info(f"{len(uniq_ants)} unique antenna: {uniq_ants}")
        logger.info(f"Unique fields: {uniq_fields}")
        logger.info(f"Phase direction: {phase_dir}")

    return MSSummary(
        flagged=flagged,
        unflagged=unflagged,
        flag_spectrum=flag_spectrum,
        fields=uniq_fields,
        ants=uniq_ants,
        beam=beam_no,
        path=ms.path,
        phase_dir=phase_dir,
    )


def split_by_field(
    ms: MS | Path,
    field: str | None = None,
    out_dir: Path | None = None,
    column: str | None = None,
) -> list[MS]:
    """Attempt to split an input measurement set up by the unique FIELDs recorded

    Args:
        ms (Union[MS, Path]): Input measurement sett to split into smaller MSs by field name
        field (Optional[str], optional): Desired field to extract. If None, all are split. Defaults to None.
        out_dir (Optional[Path], optional): Output directory to write the fresh MSs to. If None, write to same directory as
        parent MS. Defaults to None.
        column (Optional[str], optional): If not None, set the column attribute of the output MS instance to this. Defaults to None.

    Returns:
        List[MS]: The output MSs split by their field name.
    """
    ms = MS.cast(ms)

    # TODO: Split describe_ms up so can get just fields
    ms_summary = describe_ms(ms, verbose=False)

    logger.info("Collecting field names and corresponding FIELD_IDs")
    fields = [field] if field else ms_summary.fields
    field_idxs = [get_field_id_for_field(ms=ms, field_name=field) for field in fields]

    out_mss: list[MS] = []

    ms_out_dir: Path = Path(out_dir) if out_dir is not None else ms.path.parent
    logger.info(f"Will write output MSs to {ms_out_dir}.")

    if not ms_out_dir.exists():
        try:
            logger.info(f"Creating {ms_out_dir}.")
            ms_out_dir.mkdir(parents=True)
        except Exception as e:
            logger.warning(e)
            pass  # In case above fails due to race condition

    logger.info(f"Opening {ms.path}. ")
    with table(str(ms.path), ack=False) as tab:  # noqa: F841
        for split_name, split_idx in zip(fields, field_idxs):
            logger.info(f"Selecting FIELD={split_name}")
            sub_ms = taql(f"select * from $tab where FIELD_ID=={split_idx}")

            out_ms_str = create_ms_name(ms_path=ms.path, field=split_name)
            out_path = ms_out_dir / Path(out_ms_str).name

            logger.info(f"Writing {out_path!s} for {split_name}")
            sub_ms.copy(str(out_path), deep=True)

            out_mss.append(
                MS(path=out_path, beam=get_beam_from_ms(out_path), column=column)
            )

    return out_mss


def check_column_in_ms(
    ms: MS | str | PathLike,
    column: str | None = None,
    sub_table: str | None = None,
) -> bool:
    """Checks to see whether a column exists in an MS. If `column` is provided this
    is checked. It `column` is None, then the MS.column is specified. If both are
    `None` then an error is raised.

    Args:
        ms (Union[MS, str, PathLike]): The measurement set to check. Will attempt to cast to Path.
        column (Optional[str], optional): The column to check for. Defaults to None. sub_table (Optional[str], optional): A sub-table of the measurement set to inspect. If `None` the main table is examined. Defaults to None.

    Raises:
        ValueError: Raised when both `column` and `ms.column` are None.

    Returns:
        bool: Whether the column exists in the measurement set.
    """

    check_col = column
    if isinstance(ms, MS):
        logger.debug(f"{ms.column=} {column=}")
        check_col = column if column is not None else ms.column

    if check_col is None:
        raise ValueError(f"No column to check specified: {ms} {column=}.")

    ms_path = ms.path if isinstance(ms, MS) else Path(ms)
    check_table = str(ms_path) if sub_table is None else f"{ms_path!s}/{sub_table}"

    logger.debug(f"Checking for {check_col} in {check_table}")
    with table(check_table, readonly=True) as tab:
        tab_cols = tab.colnames()
        logger.debug(f"{ms_path} contains columns={tab_cols}.")
        result = check_col in tab_cols

    return result


def consistent_ms(ms1: MS, ms2: MS) -> bool:
    """Perform some basic consistency checks to ensure MS1 can
    be combined with MS2. This is important when considering
    candidate AO-style calibration solution files.

    Args:
        ms1 (MS): The first MS to consider
        ms2 (MS): The second MS to consider

    Returns:
        bool: Whether MS1 is consistent with MS2
    """

    logger.info(f"Comparing ms1={ms1.path!s} to ms2={(ms2.path)}")
    beam1 = get_beam_from_ms(ms=ms1)
    beam2 = get_beam_from_ms(ms=ms2)

    result = True
    reasons = []
    if beam1 != beam2:
        logger.debug(f"Beams are different: {beam1=} {beam2=}")
        reasons.append(f"{beam1=} != {beam2=}")
        result = False

    freqs1 = get_freqs_from_ms(ms=ms1)
    freqs2 = get_freqs_from_ms(ms=ms2)

    if len(freqs1) != len(freqs2):
        logger.debug(f"Length of frequencies differ: {len(freqs1)=} {len(freqs2)=}")
        reasons.append(f"{len(freqs1)=} != {len(freqs2)=}")
        result = False

    min_freqs1, min_freqs2 = np.min(freqs1), np.min(freqs2)
    if min_freqs1 != min_freqs2:
        logger.debug(f"Minimum frequency differ: {min_freqs1=} {min_freqs2=}")
        reasons.append(f"{min_freqs1=} != {min_freqs2=}")
        result = False

    max_freqs1, max_freqs2 = np.max(freqs1), np.max(freqs2)
    if min_freqs1 != min_freqs2:
        logger.debug(f"Maximum frequency differ: {max_freqs1=} {max_freqs2=}")
        reasons.append(f"{max_freqs1=} != {max_freqs2=}")
        result = False

    if not result:
        logger.info(f"{ms1.path!s} not compatibale with {ms2.path!s}, {reasons=}")

    return result


def consistent_channelwise_frequencies(
    freqs: list[np.ndarray] | np.ndarray,
) -> np.ndarray:
    """Given a collection of frequencies in the form of
    (N, frequencies), inspect the frequencies channelwise
    to ensure they are all the same.

    This does not operate on MSs, just the collection of frequencies

    Args:
        freqs (Union[List[np.ndarray], np.ndarray]): The collection of frequencies to be inspected

    Returns:
        np.ndarray: Same length as the frequencies. True if for a single channel all frequencies are the same. False otherwise.
    """
    freqs = np.array(freqs)
    assert len(freqs.shape) == 2, (
        f"{freqs.shape=}, but was expecting something of rank 2"
    )

    freqs_are_same = np.all(freqs - freqs[0, None] == 0, axis=1)
    assert len(freqs_are_same.shape) == 1, (
        f"Channelwise check should be length 1, but have {freqs_are_same.shaope=}"
    )
    return freqs_are_same


def consistent_ms_frequencies(mss: tuple[MS, ...]) -> bool:
    """Given a set of measurement sets, inspect the frequencies
    to ensure they are all the same

    See the ``get_freqs_from_ms`` function, which is used
    internally.

    Args:
        mss (Tuple[MS, ...]): Collection of MSs to inspect the frequencies of

    Returns:
        bool: Whether all the frequencies and their order are the same
    """

    logger.info(f"Collection frequencies from {len(mss)} measurement sets")
    freqs = [get_freqs_from_ms(ms=ms) for ms in mss]

    all_the_same = consistent_channelwise_frequencies(freqs=freqs)

    return np.all(all_the_same)


def rename_column_in_ms(
    ms: MS,
    original_column_name: str,
    new_column_name: str,
    update_tracked_column: bool = False,
) -> MS:
    """Rename a column in a measurement set. Optionally update the tracked
    `data` column attribute of the input measurement set.

    Args:
        ms (MS): Measurement set with the column to rename
        original_column_name (str): The name of the column that will be changed
        new_column_name (str): The new name of the column set in `original_column_name`
        update_tracked_column (bool, optional): Whether the `data` attribute of `ms` will be updated to `new_column_name`. Defaults to False.

    Returns:
        MS: The measurement set operated on
    """
    ms = MS.cast(ms=ms)

    with table(tablename=str(ms.path), readonly=False, ack=False) as tab:
        colnames = tab.colnames()
        assert original_column_name in colnames, (
            f"{original_column_name=} missing from {ms}"
        )
        assert new_column_name not in colnames, (
            f"{new_column_name=} already exists in {ms}"
        )

        logger.info(f"Renaming {original_column_name} to {new_column_name}")
        tab.renamecol(oldname=original_column_name, newname=new_column_name)
        tab.flush()

    if update_tracked_column:
        ms = ms.with_options(column=new_column_name)

    return ms


def remove_columns_from_ms(
    ms: MS | Path, columns_to_remove: str | list[str]
) -> list[str]:
    """Attempt to remove a collection of columns from a measurement set.
    If any of the provided columns do not exist they are ignored.

    Args:
        ms (Union[MS, Path]): The measurement set to inspect and remove columns from
        columns_to_remove (Union[str, List[str]]): Collection of column names to remove. If a single column internally it is cast to a list of length 1.

    Returns:
        List[str]: Collection of column names removed
    """

    if isinstance(columns_to_remove, str):
        columns_to_remove = [columns_to_remove]

    ms = MS.cast(ms=ms)
    with table(tablename=str(ms.path), readonly=False, ack=False) as tab:
        colnames = tab.colnames()
        columns_to_remove = [c for c in columns_to_remove if c in colnames]
        if len(columns_to_remove) == 0:
            logger.info(f"All columns provided do not exist in {ms.path}")
        else:
            logger.info(f"Removing {columns_to_remove=} from {ms.path}")
            tab.removecols(columnnames=columns_to_remove)

    return columns_to_remove


def subtract_model_from_data_column(
    ms: MS,
    model_column: str = "MODEL_DATA",
    data_column: str | None = None,
    output_column: str | None = None,
    update_tracked_column: bool = False,
) -> MS:
    """Execute a ``taql`` query to subtract the MODEL_DATA from a nominated data column.
    This requires the ``model_column`` to already be inserted into the MS. Internally
    the ``critical_ms_interaction`` context manager is used to highlight that the MS
    is being modified should things fail when subtracting.

    Args:
        ms (MS): The measurement set instance being considered
        model_column (str, optional): The column with representing the model. Defaults to "MODEL_DATA".
        data_column (Optional[str], optional): The column where the column will be subtracted. If ``None`` it is taken from the ``column`` nominated by the input ``MS`` instance. Defaults to None.
        output_column (Optional[str], optional): The output column that will be created. If ``None`` it defaults to ``data_column``. Defaults to None.
        update_tracked_column (bool, optional): If True, update ``ms.column`` to the column with subtracted data. Defaults to False.

    Returns:
        MS: The updated MS
    """
    ms = MS.cast(ms)
    data_column = data_column if data_column else ms.column
    assert data_column is not None, f"{data_column=}, which is not allowed"

    output_column = output_column if output_column else data_column
    assert output_column is not None, f"{output_column=}, which is not allowed"
    with critical_ms_interaction(input_ms=ms.path) as critical_ms:
        with table(str(critical_ms), readonly=False) as tab:
            logger.info("Extracting columns")
            colnames = tab.colnames()
            assert all([d in colnames for d in (model_column, data_column)]), (
                f"{model_column=} or {data_column=} missing from {colnames=}"
            )

            if output_column not in colnames:
                from casacore.tables import makecoldesc

                logger.info(f"Adding {output_column=}")
                desc = makecoldesc(data_column, tab.getcoldesc(data_column))
                desc["name"] = output_column
                tab.addcols(desc)
                tab.flush()

            logger.info(f"Subtracting {model_column=} from {data_column=}")
            taql(f"UPDATE $tab SET {output_column}={data_column}-{model_column}")

    if update_tracked_column:
        logger.info(f"Updating ms.column to {output_column=}")
        ms = ms.with_options(column=output_column)
    return ms


# TODO: Clean up the usage and description of the argument `instrument_column`
# as it is currently being used in unclear ways. Specifically there is a renaming
# of the data_column to instrument_column before the rotation of things
def preprocess_askap_ms(
    ms: MS | Path,
    data_column: str = "DATA",
    instrument_column: str = "INSTRUMENT_DATA",
    overwrite: bool = True,
    skip_rotation: bool = False,
    fix_stokes_factor: bool = False,
    apply_ms_transform: bool = False,
    casa_container: Path | None = None,
    rename: bool = False,
) -> MS:
    """The ASKAP MS stores its data in a way that is not immediately accessible
    to other astronomical software, like wsclean or casa. For each measurement set
    the centre of the field is stored, and beam offsets are stored in a separate table.

    Additionally, the correlations stored are more akin to (P, Q) -- they are not
    (X, Y) in the sky reference frame. This function does two things:

    * updates the positions stored so when data are imaged/calibrated the correlations are directed to the correct position
    * will apply a rotation to go from (P, Q) -> (X, Y)

    These corrections are applied to the original MS, and should be
    able to be executed multiple times without accumulating changes.

    Args:
        ms (Union[MS, Path]): The measurement set to update
        data_column (str, optional): The name of the data column to correct. This will first be renamed to the value specified by `instrument_column` before being corrected. Defaults to 'DATA'.
        instrument_column (str, optional): The name of the column that will hold the original `data_column` data. Defaults to 'INSTRUMENT_DATA'
        overwrite (bool, optional): If the `instrument_column` and `data_column` both exist and `overwrite=True` the `data_column` will be overwritten. Otherwise, a `ValueError` is raised. Defaults to True.
        skip_rotation (bool, optional): If true, the visibilities are not rotated Defaults to False.
        fix_stokes_factor (bool, optional): Apply the stokes scaling factor (aruses in different definition of Stokes between Ynadasoft and other applications) when rotation visibilities. This should be set to False is the bandpass solutions have already absorded this scaling term. Defaults to False.
        apply_ms_transform (bool, optional): If True, the MS will be transformed using the `casa_container` provided. Defaults to False.
        casa_container (Path, optional): The path to the CASA container that will be used to transform the MS. Defaults to None.
        rename (bool, optional): If True, the MS will be renamed to a Flint-processed name. Defaults to False.

    Returns:
        MS: An updated measurement set with the corrections applied.
    """
    ms = MS.cast(ms)

    assert data_column != instrument_column, (
        f"Received matching column names: {data_column=} {instrument_column=}"
    )

    if apply_ms_transform:
        if casa_container is None:
            raise ValueError(
                "apply_ms_transform=True, but no casa_container provided. "
            )
        corrected_ms_path = copy_with_mstranform(
            casa_container=casa_container,
            ms_path=ms.path,
        )
        ms = MS.cast(corrected_ms_path)

    logger.info(f"Will be running ASKAP MS conversion operations against {ms.path!s}.")
    logger.info("Correcting directions. ")

    with table(str(ms.path), ack=False, readonly=False) as tab:
        colnames = tab.colnames()
        if data_column not in colnames:
            raise ValueError(
                f"Column {data_column} not found in {ms.path!s}. Columns found: {colnames}"
            )
        if all([col in colnames for col in (data_column, instrument_column)]):
            msg = (
                f"Column {instrument_column} already in {ms.path!s}. Already corrected?"
            )
            if not overwrite:
                raise ValueError(msg)

        if (
            not skip_rotation
            and data_column in colnames
            and instrument_column not in colnames
        ):
            logger.info(f"Renaming {data_column} to {instrument_column}.")
            tab.renamecol(data_column, instrument_column)

        tab.flush()

    logger.info("Correcting the field table. ")
    fix_ms_dir(ms=str(ms.path))

    if skip_rotation:
        # TODO: Should we copy the DATA to INSTRUMENT_DATA?
        logger.info("Skipping the rotation of the visibilities. ")
        ms = ms.with_options(column=data_column)
        logger.info(f"Returning {ms=}.")
        return ms

    logger.info("Applying rotation matrix to correlations. ")
    logger.info(
        f"Rotating visibilities for {ms.path} with data_column={instrument_column} and corrected_data_column={data_column}"
    )
    fix_ms_corrs(
        ms=ms.path,
        data_column=instrument_column,
        corrected_data_column=data_column,
        fix_stokes_factor=fix_stokes_factor,
    )

    if rename:
        logger.info("Renaming the measurement set.")
        new_ms_name_str = create_ms_name(ms_path=ms.path)
        new_ms_path = ms.path.parent / Path(new_ms_name_str)
        shutil.move(ms.path, new_ms_path)
        ms = ms.with_options(path=new_ms_path)

    return ms.with_options(column=data_column)


def copy_and_preprocess_casda_askap_ms(
    casda_ms: MS | Path,
    data_column: str = "DATA",
    instrument_column: str = "INSTRUMENT_DATA",
    fix_stokes_factor: bool = True,
    output_directory: Path = Path("./"),
) -> MS:
    """Convert an ASKAP pipeline MS from CASDA into a FLINT form. This involves
    making a copy of it, updating its name, and then preprocessing.

    Typically these MSs are:

    - bandpass calibrated
    - self-calibrated
    - in the instrument frame
    - has factor of 2 scaling

    This function attempts to craft the MS so that the data column has had visibilities rotated
    and scaled to make them compatible with certain imaging packages (e.g. wsclean).

    Args:
        casda_ms (Union[MS, Path]): The measurement set to preprocess
        data_column (str, optional): The column with data to preprocess. Defaults to "DATA".
        instrument_column (str, optional): The name of the column to be created with data in the instrument frame. Defaults to "INSTRUMENT_DATA".
        fix_stokes_factor (bool, optional): Whether to scale the visibilities to account for the factor of 2 error. Defaults to True.
        output_directory (Path, optional): The output directory that the preprocessed MS will be placed into. Defaults to Path("./").

    Returns:
        MS: a corrected and preprocessed measurement set
    """
    ms = MS.cast(casda_ms)

    out_ms_path = output_directory / create_ms_name(ms_path=ms.path)
    logger.info(f"New MS name: {out_ms_path}")
    out_ms_path = copy_directory(input_directory=ms.path, output_directory=out_ms_path)

    ms = ms.with_options(path=out_ms_path)

    logger.info(
        f"Will be running CASDA ASKAP MS conversion operations against {ms.path!s}."
    )

    with table(str(ms.path), ack=False, readonly=False) as tab:
        column_names = tab.colnames()
        assert data_column in column_names and instrument_column not in column_names, (
            f"{ms.path} column names failed. {data_column=} {instrument_column=} {column_names=}"
        )
        tab.renamecol(data_column, instrument_column)
        tab.flush()

    logger.info("Correcting directions. ")
    fix_ms_dir(ms=str(ms.path))

    logger.info("Applying rotation matrix to correlations. ")
    logger.info(
        f"Rotating visibilities for {ms.path} with data_column={instrument_column} and corrected_data_column={data_column}"
    )
    fix_ms_corrs(
        ms=ms.path,
        data_column=instrument_column,
        corrected_data_column=data_column,
        fix_stokes_factor=fix_stokes_factor,
    )

    return ms.with_options(column=data_column)


def rename_ms_and_columns_for_selfcal(
    ms: MS,
    target: str | Path,
    corrected_data: str = "CORRECTED_DATA",
    data: str = "DATA",
) -> MS:
    """Take an existing measurement set, rename it, and appropriately
    rename the "DATA" and "CORRECTED_DATA" columns to support a new
    round of imaging and self-calibration.

    This could be considered for larger measurement sets where holding
    multiple copies throughout rounds of self-calibration is not advisable.

    Args:
        ms (MS): The subject measurement set to rename
        target (Union[str, Path]): The targett path the measurement set will be renamed to. This should not already exist.
        corrected_data (str, optional): The name of the column with the latest calibrated data. This becomes the `data` column. Defaults to "CORRECTED_DATA".
        data (str, optional): The name of the column that will be subsequently processed. If it exists it will be removed. Defaults to "DATA".

    Raises:
        FileExistsError: Raised if `target` already exists
        FileNotFoundError: Raise if `ms.path` does not exist

    Returns:
        MS: Updated MS container with new path and appropriate data column
    """

    ms = MS.cast(ms)
    target = Path(target)

    if target.exists():
        raise FileExistsError(f"{target} already exists!")
    if not ms.path.exists() or not ms.path.is_dir():
        raise FileNotFoundError(
            f"{ms.path} does not exists or is not a directory (hence measurement set)."
        )

    logger.info(f"Renaming {ms.path} to {target=}")
    ms.path.rename(target=target)

    # Just some sanity in case None is passed through
    if not (corrected_data or data):
        return ms.with_options(path=target)

    # Now move the corrected data column into the column to be imaged.
    # For casa it really needs to be DATA
    with table(str(target), readonly=False, ack=False) as tab:
        colnames = tab.colnames()

        if (
            all([col in colnames for col in (corrected_data, data)])
            and corrected_data != data
        ):
            logger.info(f"Removing {data} and renaming {corrected_data}")
            tab.removecols(columnnames=data)
            tab.renamecol(oldname=corrected_data, newname=data)
        elif (
            corrected_data in colnames
            and data not in controlnames
            and corrected_data != data
        ):
            logger.info(f"Renaming {corrected_data=} to {data=}")
            tab.renamecol(oldname=corrected_data, newname=data)
        elif corrected_data not in colnames and data in colnames:
            logger.warning(f"No {corrected_data=}, and {data=} seems to be present")
        elif (
            all([col in colnames for col in (corrected_data, data)])
            and corrected_data == data
            and corrected_data != "DATA"
        ):
            data = "DATA"
            logger.info(f"Renaming {corrected_data} to DATA")
            tab.renamecol(corrected_data, data)

        tab.flush()

    # This is a safe guard against my bad handling of the above / mutineers
    # There could be interplay with these columns when potato peel is used
    # as some MSs will have CORRECYED_DATA and others may not.
    assert data == "DATA", (
        f"Somehow data column is not DATA, instead {data=}. Likely a problem for casa."
    )

    return ms.with_options(path=target, column=data)


def find_mss(
    mss_parent_path: Path,
    expected_ms_count: int | None = 36,
    data_column: str | None = None,
    model_column: str | None = None,
) -> tuple[MS, ...]:
    """Search a directory to find measurement sets via a simple
    `*.ms` glob expression. An expected number of MSs can be enforced
    via the `expected_ms_count` option.

    Args:
        mss_parent_path (Path): The parent directory that will be globbed to search for MSs.
        expected_ms_count (Optional[int], optional): The number of MSs that should be there. If None no check is performed. Defaults to 36.
        data_column (Optional[str], optional): Set the column attribute of each MS to this (no checks to ensure it exists). If None use default of MS. Defaults to None.
        model_column (Optional[str], optional): Set the model column attribute of each MS to this (no checks to ensure it exists). If None use default of MS. Defaults to None.

    Returns:
        Tuple[MS, ...]: Collection of found MSs
    """
    assert mss_parent_path.exists() and mss_parent_path.is_dir(), (
        f"{mss_parent_path!s} does not exist or is not a folder. "
    )

    found_mss = tuple(
        [MS.cast(ms_path) for ms_path in sorted(mss_parent_path.glob("*.ms"))]
    )

    if expected_ms_count:
        assert len(found_mss) == expected_ms_count, (
            f"Expected to find {expected_ms_count} in {mss_parent_path!s}, found {len(found_mss)}."
        )

    if data_column or model_column:
        logger.info(f"Updating column attribute to {data_column=}")
        found_mss = tuple(
            [
                found_ms.with_options(column=data_column, model_column=model_column)
                for found_ms in found_mss
            ]
        )

    return found_mss


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Components to interact with MS")

    subparser = parser.add_subparsers(dest="mode")

    split_parser = subparser.add_parser(
        "split", help="Split an MS based on field name. "
    )

    split_parser.add_argument("ms", type=Path, help="MS to split based on fields. ")
    split_parser.add_argument(
        "--ms-out-dir",
        type=Path,
        default=None,
        help="Location to write the output MSs to. ",
    )

    preprocess_parser = subparser.add_parser(
        "preprocess",
        help="Apply preprocessing operations to the ASKAP MS so it can be used outside of yandasoft",
    )

    preprocess_parser.add_argument("ms", type=Path, help="Measurement set to correct. ")

    compatible_parser = subparser.add_parser(
        "compatible", help="Some basic checks to ensure ms1 is consistent with ms2."
    )
    compatible_parser.add_argument(
        "ms1", type=Path, help="The first measurement set to consider. "
    )
    compatible_parser.add_argument(
        "ms2", type=Path, help="The second measurement set to consider. "
    )

    casda_parser = subparser.add_parser(
        "casda",
        help="Apply preprocessing operations to the CASDA ASKAP pipeline MS so it can be used outside of yandasoft",
    )

    casda_parser.add_argument(
        "casda_ms",
        type=Path,
        help="Path to the ASKAP pipeline produced MS obtained through casda",
    )
    casda_parser.add_argument(
        "--output-directory",
        type=Path,
        default=Path("./"),
        help="Directory to write the new FLINT MS to",
    )

    return parser


def cli() -> None:
    import logging

    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "split":
        split_by_field(ms=args.ms, out_dir=args.ms_out_dir)
    if args.mode == "preprocess":
        preprocess_askap_ms(ms=args.ms)
    if args.mode == "compatible":
        res = consistent_ms(ms1=MS(path=args.ms1), ms2=MS(path=args.ms2))
        if res:
            logger.info(f"{args.ms1} is compatible with {args.ms2}")
        else:
            logger.info(f"{args.ms1} is not compatible with {args.ms2}")
    if args.mode == "casda":
        copy_and_preprocess_casda_askap_ms(
            casda_ms=Path(args.casda_ms), output_directory=Path(args.output_directory)
        )


if __name__ == "__main__":
    cli()



"""Very basic tests to make sure the FieldOptions class is
somewhat tracked, especially when using an argparse object
to create it
"""

from __future__ import annotations

from pathlib import Path

import pytest
from pydantic.fields import FieldInfo

from flint.options import (
    FieldOptions,
    _create_argparse_options,
    create_options_from_parser,
    dump_field_options_to_yaml,
    options_to_dict,
)
from flint.prefect.flows.continuum_pipeline import get_parser


def test_fieldinfo_to_argparse_options():
    """The pydantic ``FieldInfo`` object is used to generate the options that would be
    splat into an ArgumentParser.add_argument method. Ensure the expected mappings from
    types to argument options make sense"""
    field = FieldInfo(default=1, annotation=int, description="An example description")
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_name == "--jack-sparrow"
    assert field_options["action"] == "store"
    assert field_options["default"] == 1
    assert field_options["help"] == "An example description"

    field = FieldInfo(annotation=int, description="An example description")
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_name == "jack_sparrow"
    assert field_options["action"] == "store"
    assert field_options["help"] == "An example description"

    field = FieldInfo(
        default=[1, 2, 3, 4], annotation=list[int], description="An example description"
    )
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_name == "--jack-sparrow"
    assert field_options["action"] == "store"
    assert field_options["default"] == [1, 2, 3, 4]
    assert field_options["help"] == "An example description"
    assert field_options["nargs"] == "+"

    field = FieldInfo(
        default=("foo", "bar", 3),
        annotation=tuple[str, str, int],
    )
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_options["default"] == ("foo", "bar", 3)
    assert field_options["nargs"] == 3

    field = FieldInfo(
        default=("foo", "bar"),
        annotation=tuple[str, ...],
    )
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_options["default"] == (
        "foo",
        "bar",
    )
    assert field_options["nargs"] == "+"

    field = FieldInfo(
        default=None,
        annotation=tuple[str, str] | None,
    )
    field_name, field_options = _create_argparse_options(
        name="jack_sparrow", field=field
    )
    assert field_options["default"] == None  # noqa E711
    assert field_options["nargs"] == 2


def test_options_to_dict():
    """See ifthe utility around converting Option/Results to dictionary works"""
    flagger_container = Path("a")
    calibrate_container = Path("b")
    field_options_1 = FieldOptions(
        flagger_container=flagger_container, calibrate_container=calibrate_container
    )
    field_options_2 = dict(
        flagger_container=flagger_container, calibrate_container=calibrate_container
    )

    for field_options in (field_options_1, field_options_2):
        field_dict = options_to_dict(input_options=field_options)
        assert isinstance(field_dict, dict)
        assert field_dict["flagger_container"] == flagger_container
        assert field_dict["calibrate_container"] == calibrate_container

    with pytest.raises(TypeError):
        _ = options_to_dict(input_options="Jack")


def test_dump_field_options_to_yaml(tmpdir):
    """See if the field options file can be dumped to an output directory"""
    tmpdir = Path(tmpdir)

    field_options = FieldOptions(
        flagger_container=Path("a"), calibrate_container=Path("b")
    )

    assert not (tmpdir / "Jack").exists()

    path_1 = tmpdir / "field_options.yaml"
    path_2 = tmpdir / "Jack" / "Sparrow" / "field_options.yaml"

    for path in (path_1, path_2):
        output_path = dump_field_options_to_yaml(
            output_path=path, field_options=field_options
        )
        assert output_path.exists()

    with pytest.raises(FileExistsError):
        dump_field_options_to_yaml(output_path=path_2, field_options=field_options)


def test_config_field_options(tmpdir):
    output_file = f"{tmpdir}/example.config"
    contents = """--holofile /scratch3/projects/spiceracs/RACS_Low2_Holography/akpb.iquv.square_6x6.63.887MHz.SB39549.cube.fits
        --calibrate-container /scratch3/gal16b/containers/calibrate.sif
        --flagger-container /scratch3/gal16b/containers/aoflagger.sif
        --wsclean-container /scratch3/projects/spiceracs/singularity_images/wsclean_force_mask.sif
        --yandasoft-container /scratch3/gal16b/containers/yandasoft.sif
        --cluster-config /scratch3/gal16b/split/petrichor.yaml
        --rounds 2
        --split-path $(pwd)
        --zip-ms
        --run-aegean
        --use-beam-masks
        --use-preflagger
        --aegean-container '/scratch3/gal16b/containers/aegean.sif'
        --reference-catalogue-directory '/scratch3/gal16b/reference_catalogues/'
        --linmos-residuals
    """

    with open(output_file, "w") as out:
        for line in contents.split("\n"):
            out.write(f"{line.lstrip()}\n")

    parser = get_parser()
    args = parser.parse_args(
        f"""/scratch3/gal16b/askap_sbids/112334/
        --calibrated-bandpass-path /scratch3/gal16b/askap_sbids/111/
        --cli-config {output_file!s}""".split()
    )

    field_options = create_options_from_parser(
        parser_namespace=args, options_class=FieldOptions
    )

    assert isinstance(field_options, FieldOptions)
    assert field_options.zip_ms is True
    assert field_options.linmos_residuals is True
    assert field_options.rounds == 2
    assert isinstance(field_options.wsclean_container, Path)
    assert field_options.use_beam_masks
    assert field_options.use_preflagger


def test_create_field_options():
    parser = get_parser()
    args = parser.parse_args(
        """/scratch3/gal16b/askap_sbids/112334/
        --calibrated-bandpass-path /scratch3/gal16b/askap_sbids/111/
        --holofile /scratch3/projects/spiceracs/RACS_Low2_Holography/akpb.iquv.square_6x6.63.887MHz.SB39549.cube.fits
        --calibrate-container /scratch3/gal16b/containers/calibrate.sif
        --flagger-container /scratch3/gal16b/containers/aoflagger.sif
        --wsclean-container /scratch3/projects/spiceracs/singularity_images/wsclean_force_mask.sif
        --yandasoft-container /scratch3/gal16b/containers/yandasoft.sif
        --cluster-config /scratch3/gal16b/split/petrichor.yaml
        --rounds 2
        --split-path $(pwd)
        --zip-ms
        --run-aegean
        --aegean-container '/scratch3/gal16b/containers/aegean.sif'
        --reference-catalogue-directory '/scratch3/gal16b/reference_catalogues/'
        --linmos-residuals
    """.split()
    )

    field_options = FieldOptions(
        flagger_container=args.flagger_container,
        calibrate_container=args.calibrate_container,
        holofile=args.holofile,
        expected_ms=args.expected_ms,
        wsclean_container=args.wsclean_container,
        yandasoft_container=args.yandasoft_container,
        rounds=args.rounds,
        zip_ms=args.zip_ms,
        run_aegean=args.run_aegean,
        aegean_container=args.aegean_container,
        no_imaging=args.no_imaging,
        reference_catalogue_directory=args.reference_catalogue_directory,
        linmos_residuals=args.linmos_residuals,
        beam_cutoff=args.beam_cutoff,
        pb_cutoff=args.pb_cutoff,
        use_preflagger=args.use_preflagger,
    )

    assert isinstance(field_options, FieldOptions)
    assert field_options.use_preflagger is False
    assert field_options.zip_ms is True
    assert field_options.linmos_residuals is True
    assert field_options.rounds == 2
    assert isinstance(field_options.wsclean_container, Path)


def test_create_field_options2():
    parser = get_parser()
    args = parser.parse_args(
        """/scratch3/gal16b/askap_sbids/112334/
        --calibrated-bandpass-path /scratch3/gal16b/askap_sbids/111/
        --holofile /scratch3/projects/spiceracs/RACS_Low2_Holography/akpb.iquv.square_6x6.63.887MHz.SB39549.cube.fits
        --calibrate-container /scratch3/gal16b/containers/calibrate.sif
        --flagger-container /scratch3/gal16b/containers/aoflagger.sif
        --wsclean-container /scratch3/projects/spiceracs/singularity_images/wsclean_force_mask.sif
        --yandasoft-container /scratch3/gal16b/containers/yandasoft.sif
        --cluster-config /scratch3/gal16b/split/petrichor.yaml
        --rounds 2
        --split-path $(pwd)
        --run-aegean
        --aegean-container '/scratch3/gal16b/containers/aegean.sif'
        --reference-catalogue-directory '/scratch3/gal16b/reference_catalogues/'
        --use-preflagger
    """.split()
    )

    field_options = FieldOptions(
        flagger_container=args.flagger_container,
        calibrate_container=args.calibrate_container,
        holofile=args.holofile,
        expected_ms=args.expected_ms,
        wsclean_container=args.wsclean_container,
        yandasoft_container=args.yandasoft_container,
        rounds=args.rounds,
        zip_ms=args.zip_ms,
        run_aegean=args.run_aegean,
        aegean_container=args.aegean_container,
        no_imaging=args.no_imaging,
        reference_catalogue_directory=args.reference_catalogue_directory,
        linmos_residuals=args.linmos_residuals,
        beam_cutoff=args.beam_cutoff,
        pb_cutoff=args.pb_cutoff,
        use_preflagger=args.use_preflagger,
    )

    assert isinstance(field_options, FieldOptions)
    assert field_options.use_preflagger is True
    assert field_options.zip_ms is False
    assert field_options.linmos_residuals is False
    assert field_options.rounds == 2
    assert isinstance(field_options.wsclean_container, Path)


def test_create_field_options3():
    """Make sure that the calibrated-bandpass-path can be left unchecked"""
    parser = get_parser()
    args = parser.parse_args(
        """/scratch3/gal16b/askap_sbids/112334/
        --holofile /scratch3/projects/spiceracs/RACS_Low2_Holography/akpb.iquv.square_6x6.63.887MHz.SB39549.cube.fits
        --calibrate-container /scratch3/gal16b/containers/calibrate.sif
        --flagger-container /scratch3/gal16b/containers/aoflagger.sif
        --wsclean-container /scratch3/projects/spiceracs/singularity_images/wsclean_force_mask.sif
        --yandasoft-container /scratch3/gal16b/containers/yandasoft.sif
        --cluster-config /scratch3/gal16b/split/petrichor.yaml
        --rounds 2
        --split-path $(pwd)
        --run-aegean
        --aegean-container '/scratch3/gal16b/containers/aegean.sif'
        --reference-catalogue-directory '/scratch3/gal16b/reference_catalogues/'
        --use-preflagger
    """.split()
    )

    field_options = FieldOptions(
        flagger_container=args.flagger_container,
        calibrate_container=args.calibrate_container,
        holofile=args.holofile,
        expected_ms=args.expected_ms,
        wsclean_container=args.wsclean_container,
        yandasoft_container=args.yandasoft_container,
        rounds=args.rounds,
        zip_ms=args.zip_ms,
        run_aegean=args.run_aegean,
        aegean_container=args.aegean_container,
        no_imaging=args.no_imaging,
        reference_catalogue_directory=args.reference_catalogue_directory,
        linmos_residuals=args.linmos_residuals,
        beam_cutoff=args.beam_cutoff,
        pb_cutoff=args.pb_cutoff,
        use_preflagger=args.use_preflagger,
    )

    assert isinstance(field_options, FieldOptions)
    assert field_options.use_preflagger is True
    assert field_options.zip_ms is False
    assert field_options.linmos_residuals is False
    assert field_options.rounds == 2
    assert isinstance(field_options.wsclean_container, Path)
    assert args.calibrated_bandpass_path is None



"""Some utility functions around the creation of Prefect task funners.

For this work we will be using Dask backed workers to perform the compute
operations.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml
from prefect_dask import DaskTaskRunner

from flint.utils import get_packaged_resource_path


def get_cluster_spec(cluster: str | Path) -> dict[Any, Any]:
    """
    Given a cluster name, obtain the appropriate SLURM configuration
    file appropriate for use with SLURMCluster.

    This cluster spec is expected to be consistent with the cluster_class
    and cluster_kwargs parameters that are used by dask_jobqueue based
    specifications.

    Args:
        cluster (Union[str,Path]): Name of cluster or path to a configuration to look up for processing

    Raises:
        ValueError: Raised when cluster is not in KNOWN_CLUSTERS and has not corresponding YAML file.

    Returns:
        dict[Any, Any]: Dictionary of know options/parameters for dask_jobqueue.SLURMCluster
    """

    yaml_file = None

    if Path(cluster).exists():
        yaml_file = cluster
    else:
        yaml_file = get_packaged_resource_path(
            package="flint.data.cluster_configs", filename=f"{cluster}.yaml"
        )

    if yaml_file is None or not Path(yaml_file).exists():
        raise ValueError(
            f"{cluster=} is not known, or its YAML file could not be loaded."
        )

    with open(yaml_file) as in_file:
        spec = yaml.load(in_file, Loader=yaml.Loader)

    return spec


def get_dask_runner(
    cluster: str | Path = "galaxy_small",
    extra_cluster_kwargs: dict[str, Any] | None = None,
) -> DaskTaskRunner:
    """Creates and returns a DaskTaskRunner configured to established a SLURMCluster instance
    to manage a set of dask-workers. The SLURMCluster is currently configured only for Galaxy.

    Keyword Args:
        cluster (Union[str,Path]): The cluster name that will be used to search for a cluster specification file.
                       This could be the name of a known cluster, or the name of a yaml file installed
                       among the `cluster_configs` directory of the aces module.

    Returns:
        DaskTaskRunner: A dask task runner capable of being used as a task_runner for a prefect flow
    """

    spec = get_cluster_spec(cluster)

    if extra_cluster_kwargs is not None:
        spec["cluster_kwargs"].update(extra_cluster_kwargs)

    task_runner = DaskTaskRunner(**spec)

    return task_runner



"""Tests around the casa self-calibration tooling"""

from __future__ import annotations

from pathlib import Path

from flint.casa import args_to_casa_task_string


def test_args_to_casa_task_str():
    """Ensure we can transform casa style function calls to strings"""

    transform = args_to_casa_task_string(
        task="mstransform",
        regridms=True,
        nspw=1,
        mode="channel",
        nchan=-1,
    )
    assert isinstance(transform, str)

    expected = "casa -c mstransform(regridms=True,nspw=1,mode='channel',nchan=-1)"
    assert transform == expected


def test_args_to_casa_task_str_arg_list():
    """Same as above but should an argument be a list of paths (for example) needs help"""

    paths = [
        Path("/jack/dataset1.ms"),
        Path("/jack/dataset2.ms"),
    ]
    applycal = args_to_casa_task_string(
        task="applycal",
        vis=Path("/some/other/ship/visibility.ms"),
        gaintable=paths,
    )

    expected = "casa -c applycal(vis='/some/other/ship/visibility.ms',gaintable=('/jack/dataset1.ms','/jack/dataset2.ms'))"
    assert expected == applycal



"""Tests around BANE and aegean, which are one of the source finding
tools used in flint. BANE is also used to derive signal maps that ultimately
feed the clean mask creation.
"""

from __future__ import annotations

from pathlib import Path

import pytest

from flint.exceptions import AttemptRerunException
from flint.source_finding.aegean import (
    AegeanOptions,
    BANEOptions,
    _bane_output_callback,
    _get_aegean_command,
    _get_bane_command,
    get_parser,
)


def test_get_parser():
    """Make sure that there are no conflicting options when creating the parser object"""
    _ = get_parser()


def test_bane_deadlock_callback():
    """Noticed that BANE sometimes goes into a 'deadlock' state, and will
    seemingly always issue an log (from scipy, presumably, arrr)
    around interpolation not being strictly ascending"""

    lines = (
        "30146:INFO using 8 cores",
        "30146:INFO using 7 stripes",
        "30343:WARNING The points in dimension 0 must be strictly ascending or descending",
    )
    _bane_output_callback(line=lines[0])
    _bane_output_callback(line=lines[1])

    with pytest.raises(AttemptRerunException):
        _bane_output_callback(line=lines[2])


def test_bane_options():
    bane_opts = BANEOptions(box_size=(2, 1), grid_size=(20, 10))

    assert isinstance(bane_opts, BANEOptions)

    bane_str = _get_bane_command(
        image=Path("this/is/a/test.fits"), bane_options=bane_opts
    )

    assert isinstance(bane_str, str)

    expected = f"BANE this/is/a/test.fits --cores {bane_opts.cores} --stripes {bane_opts.cores - 1} --grid 20 10 --box 2 1"
    assert expected == bane_str


def test_bane_options_with_defaults():
    bane_opts = BANEOptions(box_size=None, grid_size=None)

    assert isinstance(bane_opts, BANEOptions)

    bane_str = _get_bane_command(
        image=Path("this/is/a/test.fits"), bane_options=bane_opts
    )

    assert isinstance(bane_str, str)

    expected = f"BANE this/is/a/test.fits --cores {bane_opts.cores} --stripes {bane_opts.cores - 1}"
    assert expected == bane_str

    bane_opts = BANEOptions(box_size=(3, 5), grid_size=None)
    bane_str = _get_bane_command(
        image=Path("this/is/a/test.fits"), bane_options=bane_opts
    )
    expected = f"BANE this/is/a/test.fits --cores {bane_opts.cores} --stripes {bane_opts.cores - 1} --box 3 5"
    assert expected == bane_str


def test_aegean_options():
    """Testing basic aegean options creation and expected command"""
    aegean_options = AegeanOptions(nocov=True, maxsummits=4, autoload=True)

    # This are silly tests
    assert isinstance(aegean_options, AegeanOptions)
    assert aegean_options.maxsummits == 4

    ex_path = Path("this/is/a/test.fits")
    aegean_command = _get_aegean_command(
        image=ex_path, base_output="example", aegean_options=aegean_options
    )

    expected_cmd = "aegean this/is/a/test.fits --autoload --nocov --maxsummits 4 --table example.fits"
    assert aegean_command == expected_cmd

    aegean_options2 = AegeanOptions(nocov=False, maxsummits=40, autoload=False)
    aegean_command2 = _get_aegean_command(
        image=ex_path, base_output="example", aegean_options=aegean_options2
    )

    expected_cmd2 = "aegean this/is/a/test.fits --maxsummits 40 --table example.fits"
    assert aegean_command2 == expected_cmd2



from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from astropy.io import fits

from flint.masking import (
    MaskingOptions,
    _create_signal_from_rmsbkg,
    _minimum_absolute_clip,
    _need_to_make_signal,
    _verify_set_positive_seed_clip,
    beam_shape_erode,
    consider_beam_mask_round,
    create_beam_mask_kernel,
    create_options_from_parser,
    create_snr_mask_from_fits,
    get_parser,
    minimum_absolute_clip,
)
from flint.naming import FITSMaskNames

SHAPE = (100, 100)


def test_adaptive_minimum_absolute_clip():
    """Activate the adaptive mode of the mac"""

    image = np.ones((2000, 2000)) * -1.2
    image[100:200, 100:200] = 1.0
    mask = minimum_absolute_clip(
        image=image,
        box_size=50,
        increase_factor=1.1,
        adaptive_max_depth=1,
        adaptive_box_step=4.0,
        adaptive_skew_delta=0.2,
    )
    assert np.all(~mask)

    image = np.ones((2000, 2000)) * -1.2
    image[100:200, 100:200] = 1.0
    image[110, 130] = 2000
    mask = minimum_absolute_clip(
        image=image,
        box_size=50,
        increase_factor=1.1,
        adaptive_max_depth=1,
        adaptive_box_step=4.0,
        adaptive_skew_delta=0.2,
    )
    assert np.sum(mask) == 1


def test_private_minimum_absolute_clip():
    """The minimum absolute clipping process accepts an image
    and returns a mask.  At its heart it applies a ``minimum_filter``
    from scipy."""

    image = np.ones(SHAPE) * -1.0
    image[10, 10] = -2
    private_mbc_mask = _minimum_absolute_clip(image=image, box_size=10)
    mbc_mask = minimum_absolute_clip(image=image, box_size=10)
    assert np.all(~mbc_mask)
    assert np.all(private_mbc_mask == mbc_mask)

    image[12, 12] = 5
    private_mbc_mask = _minimum_absolute_clip(image=image)
    mbc_mask = minimum_absolute_clip(image=image)
    assert np.sum(mbc_mask) == 1
    assert np.all(private_mbc_mask == mbc_mask)

    image[12, 12] = 0
    private_mbc_mask = _minimum_absolute_clip(image=image)
    mbc_mask = minimum_absolute_clip(image=image)
    assert np.all(~mbc_mask)
    assert np.all(private_mbc_mask == mbc_mask)

    image[12, 12] = 5
    private_mbc_mask = _minimum_absolute_clip(image=image, increase_factor=3)
    mbc_mask = minimum_absolute_clip(image=image, increase_factor=3)
    assert np.all(~mbc_mask)
    assert np.all(private_mbc_mask == mbc_mask)


def test_create_signal_from_rmsbkg():
    """Make sure the operations around the creation of a signal image from
    rms / bkg images follows and handles paths vs numpy arrays"""
    shape = (100, 100)
    image = np.zeros(shape) + 10
    bkg = np.ones(shape)
    rms = np.zeros(shape) + 3

    signal = _create_signal_from_rmsbkg(image=image, rms=rms, bkg=bkg)
    assert np.all(signal == 3.0)


def test_create_signal_from_rmsbkg_with_fits(tmpdir):
    """Make sure the operations around the creation of a signal image from
    rms / bkg images follows and handles paths vs numpy arrays. Unlike the
    above this uses fits for some inputs"""
    out_path = Path(tmpdir) / "fits"
    out_path.mkdir(parents=True, exist_ok=True)

    shape = (100, 100)
    image = np.zeros(shape) + 10
    image_fits = Path(out_path) / "image.fits"
    fits.writeto(image_fits, data=image)

    bkg = np.ones(shape)
    rms = np.zeros(shape) + 3
    rms_fits = Path(out_path) / "rms.fits"
    fits.writeto(rms_fits, data=rms)

    signal = _create_signal_from_rmsbkg(image=image_fits, rms=rms_fits, bkg=bkg)
    assert np.all(signal == 3.0)


def test_need_to_make_signal():
    """Ensure the conditions around creating a signal image make sense"""
    masking_options = MaskingOptions()
    assert _need_to_make_signal(masking_options=masking_options)

    masking_options = MaskingOptions(flood_fill_use_mbc=True)
    assert not _need_to_make_signal(masking_options=masking_options)


def test_arg_parser_cli_and_masking_options():
    """See if the CLI parser is constructed with correct set of
    options which are properly converted to a MaskingOptions object"""
    parser = get_parser()
    args = parser.parse_args(
        args="mask img --rms-fits rms --bkg-fits bkg --flood-fill --flood-fill-positive-seed-clip 10 --flood-fill-positive-flood-clip 1. --flood-fill-use-mbc --flood-fill-use-mbc-box-size 100".split()
    )
    masking_options = create_options_from_parser(
        parser_namespace=args, options_class=MaskingOptions
    )
    assert isinstance(masking_options, MaskingOptions)
    assert masking_options.flood_fill
    assert masking_options.flood_fill_use_mbc
    assert masking_options.flood_fill_positive_seed_clip == 10.0
    assert masking_options.flood_fill_positive_flood_clip == 1.0


def test_create_beam_mask_kernel():
    """See whether the beam kernel creation mask can return a correct mask"""
    fits_header = fits.Header(
        dict(
            CDELT1=-0.000694444444444444,
            CDELT2=0.000694444444444444,
            BMAJ=0.00340540107886635,
            BMIN=0.00283268735470751,
            BPA=74.6618858613889,
        )
    )
    mask_1 = create_beam_mask_kernel(fits_header=fits_header)
    assert mask_1.shape == (100, 100)
    assert np.sum(mask_1) == 12

    mask_2 = create_beam_mask_kernel(fits_header=fits_header, minimum_response=0.1)
    assert mask_2.shape == (100, 100)
    assert np.sum(mask_2) == 52

    with pytest.raises(AssertionError):
        fits_header = fits.Header(
            dict(
                CDELT1=-0.000694444444444444,
                CDELT2=0.2000694444444444444,
                BMAJ=0.00340540107886635,
                BMIN=0.00283268735470751,
                BPA=74.6618858613889,
            )
        )
        create_beam_mask_kernel(fits_header=fits_header)

    with pytest.raises(KeyError):
        fits_header = fits.Header(
            dict(
                CDELT1=-0.000694444444444444,
                BMAJ=0.00340540107886635,
                BMIN=0.00283268735470751,
                BPA=74.6618858613889,
            )
        )
        create_beam_mask_kernel(fits_header=fits_header)


def test_beam_shape_erode():
    """Ensure that the beam shape erosion approach works. This should drop out pixels
    should the beam shape structure connectivity be statisifed"""
    fits_header = fits.Header(
        dict(
            CDELT1=-0.000694444444444444,
            CDELT2=0.000694444444444444,
            BMAJ=0.00340540107886635,
            BMIN=0.00283268735470751,
            BPA=74.6618858613889,
        )
    )

    mask = np.zeros((500, 500)).astype(bool)

    mask[300, 300] = True
    assert np.sum(mask) == 1

    new_mask = beam_shape_erode(mask=mask, fits_header=fits_header)
    assert np.sum(new_mask) == 0

    mask = np.zeros((200, 200)).astype(bool)
    mask[100:130, 100:130] = True
    assert np.sum(mask) == 900
    erode_mask = beam_shape_erode(mask=mask, fits_header=fits_header)
    assert np.sum(erode_mask) == 729


def test_beam_shape_erode_nobeam():
    """Ensure that the beam shape erosion approach works. This should drop out pixels
    should the beam shape structure connectivity be statisifed. This should simply return
    the input array since there is no beam information"""
    fits_header = fits.Header(
        dict(
            CDELT1=-0.000694444444444444,
            CDELT2=0.000694444444444444,
        )
    )

    mask = np.zeros((500, 500)).astype(bool)

    mask[300, 300] = True
    assert np.sum(mask) == 1
    new_mask = beam_shape_erode(mask=mask, fits_header=fits_header)
    assert new_mask is mask
    assert np.sum(new_mask) == 1


def test_consider_beam_masking_round():
    """Test to ensure the beam mask consideration log is correct"""
    lower = ("all", "ALL", "aLl")
    states = (
        consider_beam_mask_round(current_round=1, mask_rounds=low) for low in lower
    )

    assert all(states)

    assert consider_beam_mask_round(current_round=3, mask_rounds=1)
    assert not consider_beam_mask_round(current_round=0, mask_rounds=1)

    assert consider_beam_mask_round(current_round=3, mask_rounds=(1, 2, 3, 4, 5))
    assert not consider_beam_mask_round(current_round=3, mask_rounds=(1, 2, 4, 5))

    assert not consider_beam_mask_round(current_round=3, mask_rounds=None)

    assert not consider_beam_mask_round(
        current_round=3, mask_rounds=1, allow_beam_masks=False
    )
    assert consider_beam_mask_round(
        current_round=3, mask_rounds=1, allow_beam_masks=True
    )
    assert consider_beam_mask_round(current_round=1, mask_rounds=1)


@pytest.fixture
def fits_dir(tmpdir):
    fits_dir = Path(tmpdir) / "fits"
    fits_dir.mkdir()

    shape = SHAPE

    img_data = np.ones(shape)
    bkg_data = np.ones(shape) * 0.5
    rms_data = np.ones(shape) * 0.1

    fits.writeto(fits_dir / "image.fits", img_data)
    fits.writeto(fits_dir / "rms.fits", rms_data)
    fits.writeto(fits_dir / "bkg.fits", bkg_data)

    return fits_dir


def test_make_masking_options():
    """Just a dump test to make sure the options structure is ok"""

    masking_options = MaskingOptions()

    assert masking_options.base_snr_clip != -1

    masking_options = masking_options.with_options(base_snr_clip=-1)
    assert masking_options.base_snr_clip == -1


def test_verify_set_seed_clip():
    """Make sure the flood seed clip handles items above all possible values"""
    signal = np.ones((100, 100)) * 10.0

    flood_clip = _verify_set_positive_seed_clip(signal=signal, positive_seed_clip=9.0)

    assert flood_clip == 9.0
    flood_clip = _verify_set_positive_seed_clip(
        signal=signal, positive_seed_clip=999999.0
    )
    assert flood_clip == 9.0


def test_fits_masking(fits_dir):
    masking_options = MaskingOptions(flood_fill=False)
    names = create_snr_mask_from_fits(
        fits_image_path=fits_dir / "image.fits",
        fits_rms_path=fits_dir / "rms.fits",
        fits_bkg_path=fits_dir / "bkg.fits",
        masking_options=masking_options,
    )

    assert isinstance(names, FITSMaskNames)
    assert names.mask_fits.exists()
    assert names.signal_fits is None

    mask_data = fits.getdata(names.mask_fits)
    valid = np.sum(mask_data)
    assert valid == np.prod(SHAPE)


def test_fits_masking_with_signal(fits_dir):
    masking_options = MaskingOptions(flood_fill=False)
    names = create_snr_mask_from_fits(
        fits_image_path=fits_dir / "image.fits",
        fits_rms_path=fits_dir / "rms.fits",
        fits_bkg_path=fits_dir / "bkg.fits",
        masking_options=masking_options,
        create_signal_fits=True,
    )

    assert isinstance(names, FITSMaskNames)
    assert names.mask_fits.exists()
    assert names.signal_fits.exists()

    mask_data = fits.getdata(names.mask_fits)
    valid = np.sum(mask_data)
    assert valid == np.prod(SHAPE)

    signal_data = fits.getdata(names.signal_fits)
    assert np.allclose(signal_data, 5)



"""Collection of functions and tooling intended
for general usage.
"""

from __future__ import annotations

import datetime
import os
import shutil
import signal
import subprocess
from contextlib import contextmanager
from pathlib import Path
from socket import gethostname
from typing import Any, Generator, NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS

from flint.convol import BeamShape
from flint.exceptions import TimeLimitException
from flint.logging import logger

# TODO: This Captain is aware that there is a common fits getheader between
# a couple of functions that interact with tasks. Perhaps a common FITS properties
# struct should be considered. The the astropy.io.fits.Header might be
# appropriate to pass around between dask / prefect delayed functions. Something
# that only opens the FITS file once and places things into common field names.


def flatten_items(items: list[Any]) -> list[Any]:
    """Recursively flatten a collection (list or tuple) of items.

    Args:
        items (list[Any]): Items to flatten. Can be arbitrarily nested

    Returns:
        list[Any]: Flattened items
    """
    flat_items = []

    def _flatten(nested):
        for e in nested:
            if isinstance(e, (list, tuple)):
                _flatten(e)
            else:
                flat_items.append(e)

    _flatten(items)
    return flat_items


def _signal_timelimit_handler(*args):
    raise TimeLimitException


@contextmanager
def timelimit_on_context(
    timelimit_seconds: int | float,
) -> Generator[None, None, None]:
    """Creates a context manager that will raise ``flint.exceptions.TimelimitException``
    should the control not leave the ``with`` context within an specified amount of time.

    Notes:
         This function **can not** be used if the function calling it is not executing
         in the main thread, such as the case with ``dask``. The underlying ``signal``
         module relies on being in the main thead, otherwise an ``Exception`` is raised.

    Args:
        timelimit_seconds (Union[int,float]): The maximum time allowed for the with context to be escaped

    Raises:
        TimeLimitException: Raised should the maximum timelimit be violated.

    Yields:
        Generator[None, None, None]: A generating function that returns nothing
    """
    signal.signal(signal.SIGALRM, _signal_timelimit_handler)
    signal.alarm(int(timelimit_seconds))
    logger.info(f"Setting a timelimit of {timelimit_seconds=}")

    try:
        yield
    except TimeLimitException:
        logger.info(f"Timeout limit of {timelimit_seconds=} reached")
        raise TimeLimitException

    signal.alarm(0)


@contextmanager
def hold_then_move_into(
    move_directory: Path,
    hold_directory: Path | None,
    delete_hold_on_exist: bool = True,
) -> Generator[Path, None, None]:
    """Create a temporary directory such that anything within it on the
    exit of the context manager is copied over to `move_directory`.

    If `hold_directory` and `move_directory` are the same or `hold_directory` is None, then `move_directory`
    is immediately returned and no output files are copied or deleted. `move_directory` will be
    created if it does not exist.

    Args:
        move_directory (Path): Final directory location to move items into
        hold_directory (Optional[Path], optional): Location of directory to temporarily base work from. If None provided `move_directory` is returned and no copying/deleting is performed on exit. Defaults to None.
        delete_hold_on_exist (bool, optional): Whether `hold_directory` is deleted on exit of the context. Defaults to True.

    Returns:
        Path: Path to the temporary folder

    Yields:
        Iterator[Path]: Path to the temporary folder
    """
    # TODO: except extra files and folders to copy into `hold_directory` that are
    # also placed back on exit
    hold_directory = Path(hold_directory) if hold_directory else None
    move_directory = Path(move_directory)

    if hold_directory == move_directory or hold_directory is None:
        move_directory.mkdir(parents=True, exist_ok=True)
        yield move_directory
    else:
        for directory in (hold_directory, move_directory):
            if directory.exists():
                assert directory.is_dir()
            else:
                directory.mkdir(parents=True)

        assert all([d.is_dir() for d in (hold_directory, move_directory)])

        yield hold_directory

        for file_or_folder in hold_directory.glob("*"):
            logger.info(f"Moving {file_or_folder=} to {move_directory=}")
            shutil.move(str(file_or_folder), move_directory)

        if delete_hold_on_exist:
            remove_files_folders(hold_directory)


@contextmanager
def temporarily_move_into(
    subject: Path, temporary_directory: Path | None = None
) -> Generator[Path, None, None]:
    """Given a file or folder, temporarily copy it into the path specified
    by `temporary_directory` for the duration of the context manager. Upon
    exit the original copy, specified by `subject`, is removed and replaced
    by the copy within `temporary_directory`.

    `temporary_directory` will be created internally, and an error will be
    raised if it exists.

    If `temporary_directory` describes a nested path only the lowest directory
    is removed.

    If `temporary_directory` is None the `subject` path is returned and there
    is no copying and deleting performed.

    Args:
        subject (Path): The file or folder to temporarily move
        temporary_directory (Optional[Path], optional): The temporary directory to work with. If none the subject path is returned. Defaults to None.

    Yields:
        Path: The path to the temporary object
    """
    subject = Path(subject)
    temporary_directory = Path(temporary_directory) if temporary_directory else None

    if temporary_directory is None:
        yield subject
    else:
        temporary_directory.mkdir(parents=True, exist_ok=True)
        assert temporary_directory.is_dir(), (
            f"{temporary_directory=} exists and is not a folder"
        )

        output_item = temporary_directory / subject.name
        assert not output_item.exists(), f"{output_item=} already exists! "

        logger.info(f"Moving {subject=} to {output_item=}")

        if subject.is_dir():
            logger.info(f"{subject=} is a directory, recursively copying")
            copy_directory(
                input_directory=subject, output_directory=output_item.absolute()
            )
        else:
            shutil.copy(subject, output_item)

        yield output_item

        logger.info(f"Moving {output_item} back to {subject=}")
        remove_files_folders(subject)
        shutil.move(str(output_item), subject)

        logger.info(f"Removing {temporary_directory=}")
        shutil.rmtree(temporary_directory)


def parse_environment_variables(
    variable: str | None, default: str | None = None
) -> str | None:
    """Expand a $VARIABLE environment variable to obtain its value from
    the environment. The dollar-character is required in its input to be
    expanded.

    Expanding an environment variable embedded within a Path-like expression
    is supported. Each “/”-delimited segment as an env-var if it exists.

    Splits `variable` on "/", and for each segment:
      1. Strips any leading "$".
      2. If the remaining name matches an env-var, substitutes its value.
      3. If it started with "$" but the env-var is unset, returns `default`.
      4. Otherwise leaves the segment literal.

    Rejoins the segments with "/", preserves a trailing slash if present.

    Should a $VARIABLE be specified but is unresolved, the `default` value
    is returned.

    Args:
        variable: e.g. "TEST1/$SLURM_TMPDIR" or "$HOME/subdir" or "literal/path"
        default: returned if any "$VAR" lookup fails

    Returns:
        Expanded path string, `default` on lookup failure, or None if `variable` is None.
    """
    if variable is None:
        return None

    parts = variable.split("/")
    out_parts: list[str] = []

    for part in parts:
        if not part.startswith("$"):
            out_parts.append(part)
            continue

        # At this point the part should resolve to a
        # environment variable, you dirty sea dog
        name = part.lstrip("$")
        val = os.getenv(name)

        # missing placeholder → fallback
        if val is None:
            return default

        # This can not be None because of the above. Putting
        # an assert to capture this behaviour anticipating future
        # refactoring
        assert val is not None, f"{val=}, which is not expected"
        out_parts.append(val)

    result = "/".join(out_parts)

    # preserve trailing slash
    if variable.endswith("/") and not result.endswith("/"):
        result += "/"

    return result


class SlurmInfo(NamedTuple):
    hostname: str
    """The hostname of the slurm job"""
    job_id: str | None = None
    """The job ID of the slurm job"""
    task_id: str | None = None
    """The task ID of the slurm job"""
    time: str | None = None
    """The time time the job information was gathered"""


def get_slurm_info() -> SlurmInfo:
    """Collect key slurm attributes of a job

    Returns:
        SlurmInfo: Collection of slurm items from the job environment
    """

    hostname = gethostname()
    job_id = parse_environment_variables("$SLURM_JOB_ID")
    task_id = parse_environment_variables("$SLURM_ARRAY_TASK_ID")
    time = str(datetime.datetime.now())

    return SlurmInfo(hostname=hostname, job_id=job_id, task_id=task_id, time=time)


def get_job_info(mode: str = "slurm") -> SlurmInfo:
    """Get the job information for the supplied mode

    Args:
        mode (str, optional): Which mode to poll information for. Defaults to "slurm".

    Raises:
        ValueError: Raised if the mode is not supported

    Returns:
        Union[SlurmInfo]: The specified mode
    """
    # TODO: Add other modes? Return a default?
    modes = ("slurm",)

    if mode.lower() == "slurm":
        job_info = get_slurm_info()
    else:
        raise ValueError(f"{mode=} not supported. Supported {modes=} ")

    return job_info


def log_job_environment() -> SlurmInfo:
    """Log components of the slurm environment. Currently only support slurm

    Returns:
        SlurmInfo: Collection of slurm items from the job environment
    """
    # TODO: Expand this to allow potentially other job queue systems
    slurm_info = get_slurm_info()

    logger.info(f"Running on {slurm_info.hostname=}")
    logger.info(f"Slurm job id is {slurm_info.job_id}")
    logger.info(f"Slurm task id is {slurm_info.task_id}")

    return slurm_info


def get_beam_shape(fits_path: Path) -> BeamShape | None:
    """Construct and return a beam shape from the fields in a FITS image

    Args:
        fits_path (Path): FITS image to extract the beam information from

    Returns:
        Optional[BeamShape]: Shape of the beam stored in the FITS image. None is returned if the beam is not found.
    """

    header = fits.getheader(filename=fits_path)

    if not all([key in header for key in ("BMAJ", "BMIN", "BPA")]):
        return None

    beam_shape = BeamShape(
        bmaj_arcsec=header["BMAJ"] * 3600,
        bmin_arcsec=header["BMIN"] * 3600,
        bpa_deg=header["BPA"],
    )

    return beam_shape


def get_pixels_per_beam(fits_path: Path) -> float | None:
    """Given a image with beam information, return the number of pixels
    per beam. The beam is taken from the FITS header. This is evaluated
    for pixels at the reference pixel position.

    Args:
        fits_path (Path): FITS image to consideer

    Returns:
        Optional[float]: Number of pixels per beam. If beam is not in header then None is returned.
    """

    beam_shape = get_beam_shape(fits_path=fits_path)

    if beam_shape is None:
        return None

    header = fits.getheader(filename=fits_path)

    pixel_ra = np.abs(header["CDELT1"] * 3600)
    pixel_dec = np.abs(header["CDELT2"] * 3600)

    beam_area = beam_shape.bmaj_arcsec * beam_shape.bmin_arcsec * np.pi
    pixel_area = pixel_ra * pixel_dec

    no_pixels = beam_area / pixel_area

    return no_pixels


def get_packaged_resource_path(package: str, filename: str) -> Path:
    """Load in the path of a package sources.

    The `package` argument is passed as a though the module
    is being specified as an import statement: `flint.data.aoflagger`.

    Args:
        package (str): The module path to the resources
        filename (str): Filename of the datafile to load

    Returns:
        Path: The absolute path to the packaged resource file
    """
    logger.info(f"Loading {package=} for {filename=}")
    try:
        import importlib_resources as importlib_resources
    except ImportWarning:
        from importlib import resources as importlib_resources

    p = importlib_resources.files(package)
    logger.info(f"{p=}")
    full_path = Path(p) / filename  # type: ignore

    logger.debug(f"Resolved {full_path=}")

    return full_path


def generate_strict_stub_wcs_header(
    position_at_image_center: SkyCoord,
    image_shape: tuple[int, int],
    pixel_scale: u.Quantity | str,
    image_shape_is_center: bool = False,
) -> WCS:
    """Create a WCS object using some strict quantities. There
    are no attempts to cast values appropriately, exception being
    calling `astropy.units.Quantity` on the `pixel_scale` input
    should it not be a quantity.

    The supplied `image_size` is used to calculate the center of
    the image and set the reference pixel value.

    The output projection is SIN.

    Args:
        position_at_image_center (SkyCoord): The position that will be at the reference pixel
        image_shape (Tuple[int, int]): The size of the image
        pixel_scale (Union[u.Quantity,str]): Size of the square pixels. If `str` passed will be cast to `Quantity`.
        image_shape_is_center (bool, optional): It True the position specified by `image_shape` is the center reference position. if False, `image_shape` is assumed to be the size of the image, and the center is computed from this. Defaults to False.

    Raises:
        TypeError: Raised when pixel scale it not a str or astropy.units.Quantity

    Returns:
        WCS: A WCS header matching the input specs
    """

    if isinstance(pixel_scale, str):
        pixel_scale = u.Quantity(pixel_scale)
    elif not isinstance(pixel_scale, u.Quantity):
        raise TypeError(
            f"pixel_scale should be of type astro.units.Quantity or str, got {type(pixel_scale)}"
        )

    # This should be good enough
    image_center = image_shape
    if not image_shape_is_center:
        image_center = np.array(image_center) / 2
        logger.debug(f"Constructed WCS {image_center=}")

    header = {
        "CRVAL1": position_at_image_center.ra.deg,
        "CRVAL2": position_at_image_center.dec.deg,
        "CUNIT1": "deg",
        "CUNIT2": "deg",
        "CDELT1": -pixel_scale.to(u.deg).value,
        "CDELT2": pixel_scale.to(u.deg).value,
        "CRPIX1": image_center[0],
        "CRPIX2": image_center[1],
        "CTYPE1": "RA---SIN",
        "CTYPE2": "DEC--SIN",
        "SPECSYS": "TOPOCENT",
    }

    wcs = WCS(fits.Header(header))

    return wcs


def generate_stub_wcs_header(
    ra: float | u.Quantity | None = None,
    dec: float | u.Quantity | None = None,
    image_shape: tuple[int, int] | None = None,
    pixel_scale: u.Quantity | str | float | None = None,
    projection: str = "SIN",
    base_wcs: Path | WCS | None = None,
) -> WCS:
    """Create a basic WSC header object that can be used to calculate sky positions
    for an example image.

    Care should be taken when using this function as it tries to be too
    smart for its own good.

    Args:
        ra (fUnion[loat,u.Quantity]): The RA at the reference pixel. if a float is provided it is assumed to be in degrees.
        dec (Union[float,u.Quantity]): The Dec at the reference pizel. if a float is provided it is assumed to be in degrees.
        image_shape (Tuple[int, int]): Size of the representative image
        pixel_scale (Union[u.Quantity, str, float]): The size of the square pixels. if a `float` it is assumed to be arcseconds. If `str`, parsing is hangled by `astropy.units.Quantity`.
        projection (str, optional): Project scheme to encode in the header. Defaults to "SIN".
        base_wcs (Optional[Union[Path, WCS]], optional): Overload an existing WCS object with argument properties. If a `Path` the WCS is obtained from the fits file. If `None` WCS is built from arguments. Defaults to None.

    Returns:
        WCS: The representative WCS objects
    """
    # Trust nothing
    assert len(projection) == 3, (
        f"Projection should be three characters, received {projection}"
    )

    # Handle all the pixels you rotten seadog
    if pixel_scale is not None:
        if isinstance(pixel_scale, str):
            pixel_scale = u.Quantity(pixel_scale)
        elif isinstance(pixel_scale, float):
            pixel_scale = pixel_scale * u.arcsec

        # Trust nothing even more
        assert isinstance(pixel_scale, u.Quantity), (
            f"pixel_scale is not an quantity, instead {type(pixel_scale)}"
        )
        pixel_scale = np.abs(pixel_scale.to(u.rad).value)

        pixel_scale = np.array([-pixel_scale, pixel_scale])

    # Handle the ref position
    if ra is not None:
        ra = ra if isinstance(ra, u.Quantity) else ra * u.deg
    if dec is not None:
        dec = dec if isinstance(dec, u.Quantity) else dec * u.deg

    # Sort out the header. If Path get the header through and construct the WCS
    if base_wcs is not None:
        if isinstance(base_wcs, Path):
            base_wcs = WCS(fits.getheader(base_wcs)).celestial

        assert isinstance(base_wcs, WCS), (
            f"Expecting base_wcs to be a WCS object by now, instead is {type(base_wcs)}"
        )

        if image_shape is None:
            image_shape = base_wcs._naxis
        if ra is None:
            ra = base_wcs.wcs.crval[0] * u.Unit(base_wcs.wcs.cunit[0])
        if dec is None:
            dec = base_wcs.wcs.crval[1] * u.Unit(base_wcs.wcs.cunit[1])
        if pixel_scale is None:
            pixel_scale = base_wcs.wcs.cdelt

    # Only needs to be approx correct. Off by one pixel should be ok, this pirate thinks
    if image_shape is not None:
        image_center = np.array(image_shape, dtype=int) // 2

    # The celestial guarantees only two axis
    w = base_wcs.celestial if base_wcs else WCS(naxis=2)

    if any([i is None for i in (image_shape, ra, dec, pixel_scale)]):
        raise ValueError("Something is unset, and unable to form wcs object. ")

    # Nor bring it all together
    w.wcs.crpix = image_center
    w.wcs.cdelt = pixel_scale
    w.wcs.crval = [ra.to(u.deg).value, dec.to(u.deg).value]  # type: ignore
    w.wcs.ctype = [f"RA---{projection}", f"DEC--{projection}"]
    w.wcs.cunit = ["deg", "deg"]
    w._naxis = tuple(image_shape)  # type: ignore

    return w


def estimate_skycoord_centre(
    sky_positions: SkyCoord, final_frame: str = "fk5"
) -> SkyCoord:
    """Estimate the centre position of (RA, Dec) positions stored in a
    input `SkyCoord`. Internally the mean of the cartesian (X,Y,Z) is
    calculated, which is then transformed back to a sky position,.

    Args:
        sky_positions (SkyCoord): Set of input positions to consider
        final_frame (str, optional): The frame of the returned `SkyCoord` objects set using `.transform_to`. Defaults to fk5.

    Returns:
        SkyCoord: Centre position
    """
    xyz_positions = sky_positions.cartesian.xyz
    xyz_mean_position = np.mean(xyz_positions, axis=1)

    mean_position = SkyCoord(
        *xyz_mean_position, representation_type="cartesian"
    ).transform_to(final_frame)

    return mean_position


def estimate_image_centre(image_path: Path) -> SkyCoord:
    with fits.open(image_path, memmap=True) as open_image:
        image_header = open_image[0].header
        image_shape = open_image[0].data.shape

    wcs = WCS(image_header)
    centre_pixel = np.array(image_shape) / 2.0
    # The celestial deals with the radio image potentially having four dimensions
    # (stokes, frequencyes, ra, dec)
    centre_sky = wcs.celestial.pixel_to_world(centre_pixel[0], centre_pixel[1])

    return centre_sky


def zip_folder(
    in_path: Path, out_zip: Path | None = None, archive_format: str = "tar"
) -> Path:
    """Zip a directory and remove the original.

    Args:
        in_path (Path): The path that will be zipped up.
        out_zip (Path, optional): Name of the output file. A `archive_format` extension will be added by `shutil.make_archive`. Defaults to None.
        archive_format (str, optional): The format of the archive. See `shutil.make_archive`. Defaults to "tar".

    Returns:
        Path: the path of the compressed zipped folder
    """
    in_path = Path(in_path)
    out_zip = in_path if out_zip is None else out_zip

    if in_path.exists():
        logger.info(f"Zipping {in_path}.")
        shutil.make_archive(str(out_zip), format=archive_format, base_dir=str(in_path))
        remove_files_folders(in_path)
    else:
        logger.warning(f"{in_path=} does not exist... Not archiving. ")

    return out_zip


def rsync_copy_directory(target_path: Path, out_path: Path) -> Path:
    """A small attempt to rsync a directory from one location to another.
    This is an attempt to verify a copy was completed successfully.

    Args:
        target_path (Path): The target directory to copy
        out_path (Path): The location to copy the directory to

    Returns:
        Path: The output path of the new directory.
    """

    rsync_cmd = f"rsync -avh --progress --stats {target_path!s}/ {out_path!s}/ "
    logger.info(f"Rsync copying {target_path} to {out_path}.")
    logger.debug(f"Will run {rsync_cmd}")
    rsync_run = subprocess.Popen(rsync_cmd.split(), stdout=subprocess.PIPE)

    if rsync_run.stdout is not None:
        for line in rsync_run.stdout:
            logger.debug(line.decode().rstrip())

    return out_path


def copy_directory(
    input_directory: Path,
    output_directory: Path,
    verify: bool = False,
    overwrite: bool = False,
) -> Path:
    """Copy a directory into a new location.

    Args:
        input_directory (Path): The source directory to copy
        output_directory (Path): The location of the source directory to copy to
        verify (bool, optional): Attempt to run `rsync` to verify copy worked. Defaults to False.
        overwrite (bool, optional): Remove the target direcrtory if it exists. Defaults to False.

    Returns:
        Path: Location of output directory
    """

    input_directory = Path(input_directory)
    output_directory = Path(output_directory)

    assert input_directory.exists() and input_directory.is_dir(), (
        f"Currently only supports copying directories, {input_directory=} is a file or does not exist. "
    )

    logger.info(f"Copying {input_directory} to {output_directory}.")

    if output_directory.exists():
        if overwrite:
            logger.warning(f"{output_directory} already exists. Removing it. ")
            remove_files_folders(output_directory)

    shutil.copytree(input_directory, output_directory)

    if verify:
        rsync_copy_directory(input_directory, output_directory)

    return output_directory


def remove_files_folders(*paths_to_remove: Path) -> list[Path]:
    """Will remove a set of paths from the file system. If a Path points
    to a folder, it will be recursively removed. Otherwise it is simply
    unlinked.

    Args:
        paths_to_remove (Path): Set of Paths that will be removed

    Returns:
        List[Path]: Set of Paths that were removed
    """

    files_removed = []

    file: Path
    for file in paths_to_remove:
        file = Path(file)
        if not file.exists():
            logger.debug(f"{file} does not exist. Skipping, ")
            continue

        if file.is_dir():
            logger.info(f"Removing folder {file!s}")
            shutil.rmtree(file)
        else:
            logger.info(f"Removing file {file}.")
            file.unlink()

        files_removed.append(file)

    return files_removed


def create_directory(directory: Path, parents: bool = True) -> Path:
    """Will attempt to safely create a directory. Should it
    not exist it will be created. if this creates an exception,
    which might happen in a multi-process environment, it is
    reported and discarded.

    Args:
        directory (Path): Path to directory to create
        parents (bool, optional): Create parent directories if necessary. Defaults to True.

    Returns:
        Path: The directory created
    """

    directory = Path(directory)

    logger.info(f"Creating {directory!s}")
    try:
        directory.mkdir(parents=parents, exist_ok=True)
    except Exception as e:
        logger.error(f"Failed to create {directory!s} {e}.")

    return directory



"""Tests around the subtract cube imaging flow"""

from __future__ import annotations

from flint.prefect.flows.subtract_cube_pipeline import get_parser


def test_get_parser():
    """Make sure the parser can actually be builts"""
    # This is a silly one but I was bitten by it not working, arrrrhh matey
    _ = get_parser()



"""Common prefect tasks around interacting with measurement sets"""

from __future__ import annotations

from pathlib import Path
from typing import ParamSpec, TypeVar

from prefect import Task, task

from flint.imager.wsclean import WSCleanResult
from flint.logging import logger
from flint.ms import subtract_model_from_data_column
from flint.predict.addmodel import AddModelOptions, add_model

P = ParamSpec("P")
R = TypeVar("R")

task_subtract_model_from_ms = task(subtract_model_from_data_column)


# TODO: This can be a dispatcher type function should
# other modes be added
def add_model_source_list_to_ms(
    wsclean_command: WSCleanResult, calibrate_container: Path | None = None
) -> WSCleanResult:
    logger.info("Updating MODEL_DATA with source list")
    ms = wsclean_command.ms

    assert wsclean_command.image_set is not None, (
        f"{wsclean_command.image_set=}, which is not allowed"
    )

    source_list_path = wsclean_command.image_set.source_list
    if source_list_path is None:
        logger.info(f"{source_list_path=}, so not updating")
        return wsclean_command
    assert source_list_path.exists(), f"{source_list_path=} does not exist"

    if calibrate_container is None:
        logger.info(f"{calibrate_container=}, so not updating")
        return wsclean_command
    assert calibrate_container.exists(), f"{calibrate_container=} does not exist"

    add_model_options = AddModelOptions(
        model_path=source_list_path,
        ms_path=ms.path,
        mode="c",
        datacolumn="MODEL_DATA",
    )
    add_model(
        add_model_options=add_model_options,
        container=calibrate_container,
        remove_datacolumn=True,
    )
    return wsclean_command


task_add_model_source_list_to_ms: Task[P, R] = task(add_model_source_list_to_ms)
